"""
Multi purpose stuff used by both clients and servers (URI etc)

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import re
import logging
import contextlib
import ipaddress
import socket
import random
import serpent
from typing import Union, Optional
from . import config, errors, socketutil, serializers


__all__ = ["URI", "DAEMON_NAME", "NAMESERVER_NAME", "resolve", "locate_ns", "type_meta"]

log = logging.getLogger("Pyro5.core")

# standard object name for the Daemon object
DAEMON_NAME = "Pyro.Daemon"

# standard name for the Name server itself
NAMESERVER_NAME = "Pyro.NameServer"


class URI(object):
    """
    Pyro object URI (universal resource identifier).
    The uri format is like this: ``PYRO:objectid@location`` where location is one of:

    - ``hostname:port`` (tcp/ip socket on given port)
    - ``./u:sockname`` (Unix domain socket on localhost)

    There is also a 'Magic format' for simple name resolution using Name server:
      ``PYRONAME:objectname[@location]``  (optional name server location, can also omit location port)
    And one that looks up things in the name server by metadata:
      ``PYROMETA:meta1,meta2,...[@location]``  (optional name server location, can also omit location port)

    You can write the protocol in lowercase if you like (``pyro:...``) but it will
    automatically be converted to uppercase internally.
    """
    uriRegEx = re.compile(r"(?P<protocol>[Pp][Yy][Rr][Oo][a-zA-Z]*):(?P<object>\S+?)(@(?P<location>.+))?$")

    def __init__(self, uri):
        if isinstance(uri, URI):
            state = uri.__getstate__()
            self.__setstate__(state)
            return
        if not isinstance(uri, str):
            raise TypeError("uri parameter object is of wrong type")
        self.sockname = self.host = self.port = None
        match = self.uriRegEx.match(uri)
        if not match:
            raise errors.PyroError("invalid uri")
        self.protocol = match.group("protocol").upper()
        self.object = match.group("object")
        location = match.group("location")
        if self.protocol == "PYRONAME":
            self._parseLocation(location, config.NS_PORT)
        elif self.protocol == "PYRO":
            if not location:
                raise errors.PyroError("invalid uri")
            self._parseLocation(location, None)
        elif self.protocol == "PYROMETA":
            self.object = set(m.strip() for m in self.object.split(","))
            self._parseLocation(location, config.NS_PORT)
        else:
            raise errors.PyroError("invalid uri (protocol)")

    def _parseLocation(self, location, defaultPort):
        if not location:
            return
        if location.startswith("./u:"):
            self.sockname = location[4:]
            if (not self.sockname) or ':' in self.sockname:
                raise errors.PyroError("invalid uri (location)")
        else:
            if location.startswith("["):  # ipv6
                if location.startswith("[["):  # possible mistake: double-bracketing
                    raise errors.PyroError("invalid ipv6 address: enclosed in too many brackets")
                ipv6locationmatch = re.match(r"\[([0-9a-fA-F:%]+)](:(\d+))?", location)
                if not ipv6locationmatch:
                    raise errors.PyroError("invalid ipv6 address: the part between brackets must be a numeric ipv6 address")
                self.host, _, self.port = ipv6locationmatch.groups()
            else:
                self.host, _, self.port = location.partition(":")
            if not self.port:
                self.port = defaultPort
            try:
                self.port = int(self.port)
            except (ValueError, TypeError):
                raise errors.PyroError("invalid port in uri, port=" + str(self.port))

    @staticmethod
    def isUnixsockLocation(location):
        """determine if a location string is for a Unix domain socket"""
        return location.startswith("./u:")

    @property
    def location(self):
        """property containing the location string, for instance ``"servername.you.com:5555"``"""
        if self.host:
            if ":" in self.host:  # ipv6
                return "[%s]:%d" % (self.host, self.port)
            else:
                return "%s:%d" % (self.host, self.port)
        elif self.sockname:
            return "./u:" + self.sockname
        else:
            return None

    def __str__(self):
        if self.protocol == "PYROMETA":
            result = "PYROMETA:" + ",".join(self.object)
        else:
            result = self.protocol + ":" + self.object
        if self.location:
            return result + "@" + self.location
        return result

    def __repr__(self):
        return "<%s.%s at 0x%x; %s>" % (self.__class__.__module__, self.__class__.__name__, id(self), str(self))

    def __eq__(self, other):
        if not isinstance(other, URI):
            return False
        return self.__getstate__() == other.__getstate__()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__getstate__())

    def __getstate__(self):
        return self.protocol, self.object, self.sockname, self.host, self.port

    def __setstate__(self, state):
        self.protocol, self.object, self.sockname, self.host, self.port = state


class _ExceptionWrapper(object):
    """Class that wraps a remote exception. If this is returned, Pyro will
    re-throw the exception on the receiving side. Usually this is taken care of
    by a special response message flag, but in the case of batched calls this
    flag is useless and another mechanism was needed."""

    def __init__(self, exception):
        self.exception = exception

    def raiseIt(self):
        raise self.exception

    def __serialized_dict__(self):
        """serialized form as a dictionary"""
        return {
            "__class__": "Pyro5.core._ExceptionWrapper",
            "exception": serializers.SerializerBase.class_to_dict(self.exception)
        }


# register the special serializers for the pyro objects with Serpent
serpent.register_class(URI, serializers.pyro_class_serpent_serializer)
serpent.register_class(_ExceptionWrapper, serializers.pyro_class_serpent_serializer)
serializers.SerializerBase.register_class_to_dict(URI, serializers.serialize_pyro_object_to_dict, serpent_too=False)
serializers.SerializerBase.register_class_to_dict(_ExceptionWrapper, _ExceptionWrapper.__serialized_dict__, serpent_too=False)


def resolve(uri: Union[str, URI], delay_time: float = 0.0) -> URI:
    """
    Resolve a 'magic' uri (PYRONAME, PYROMETA) into the direct PYRO uri.
    It finds a name server, and use that to resolve a PYRONAME uri into the direct PYRO uri pointing to the named object.
    If uri is already a PYRO uri, it is returned unmodified.
    You can consider this a shortcut function so that you don't have to locate and use a name server proxy yourself.
    Note: if you need to resolve more than a few names, consider using the name server directly instead of repeatedly
    calling this function, to avoid the name server lookup overhead from each call.
    You can set delay_time to the maximum number of seconds you are prepared to wait until a name registration
    becomes available in the nameserver.
    """
    if isinstance(uri, str):
        uri = URI(uri)
    elif not isinstance(uri, URI):
        raise TypeError("can only resolve Pyro URIs")
    if uri.protocol == "PYRO":
        return uri
    log.debug("resolving %s", uri)
    from . import nameserver   # doing it here to avoid circular import issues
    if uri.protocol == "PYRONAME":
        with locate_ns(uri.host, uri.port) as ns:
            return nameserver.lookup(ns, uri.object, delay_time)
    elif uri.protocol == "PYROMETA":
        with locate_ns(uri.host, uri.port) as ns:
            candidates = nameserver.yplookup(ns, uri.object, None, False, delay_time)
            if candidates:
                candidate = random.choice(list(candidates.values()))
                log.debug("resolved to candidate %s", candidate)
                return URI(candidate)
            raise errors.NamingError("no registrations available with desired metadata properties %s" % uri.object)
    else:
        raise errors.PyroError("invalid uri protocol")


def locate_ns(host: Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address] = "",
              port: Optional[int] = None, broadcast: bool = True) -> "client.Proxy":
    """Get a proxy for a name server somewhere in the network."""
    from . import client
    if not host:
        # first try localhost if we have a good chance of finding it there
        if config.NS_HOST in ("localhost", "::1") or config.NS_HOST.startswith("127."):
            if ":" in config.NS_HOST:  # ipv6
                hosts = ["[%s]" % config.NS_HOST]
            else:
                # Some systems have 127.0.1.1 in the hosts file assigned to the hostname,
                # so try this too (only if it's actually used as a valid ip address)
                try:
                    socket.gethostbyaddr("127.0.1.1")
                    hosts = [config.NS_HOST] if config.NS_HOST == "127.0.1.1" else [config.NS_HOST, "127.0.1.1"]
                except socket.error:
                    hosts = [config.NS_HOST]
            for host in hosts:
                uristring = "PYRO:%s@%s:%d" % (NAMESERVER_NAME, host, port or config.NS_PORT)
                log.debug("locating the NS: %s", uristring)
                proxy = client.Proxy(uristring)
                with contextlib.suppress(errors.PyroError):
                    proxy._pyroBind()
                    log.debug("located NS")
                    return proxy
        if config.PREFER_IP_VERSION == 6:
            broadcast = False   # ipv6 doesn't have broadcast. We should probably use multicast....
        if broadcast:
            # broadcast lookup
            if not port:
                port = config.NS_BCPORT
            log.debug("broadcast locate")
            sock = socketutil.create_bc_socket(reuseaddr=config.SOCK_REUSE, timeout=0.7)
            for _ in range(3):
                try:
                    for bcaddr in config.BROADCAST_ADDRS:
                        try:
                            sock.sendto(b"GET_NSURI", 0, (bcaddr, port))
                        except socket.error as x:
                            err = getattr(x, "errno", x.args[0])
                            # handle some errno's that some platforms like to throw:
                            if err not in socketutil.ERRNO_EADDRNOTAVAIL and err not in socketutil.ERRNO_EADDRINUSE:
                                raise
                    data, _ = sock.recvfrom(100)
                    sock.close()
                    text = data.decode("iso-8859-1")
                    log.debug("located NS: %s", text)
                    proxy = client.Proxy(text)
                    return proxy
                except socket.timeout:
                    continue
            with contextlib.suppress(OSError, socket.error):
                sock.shutdown(socket.SHUT_RDWR)
            sock.close()
            log.debug("broadcast locate failed, try direct connection on NS_HOST")
        else:
            log.debug("skipping broadcast lookup")
        # broadcast failed or skipped, try PYRO directly on specific host
        host = config.NS_HOST
        port = config.NS_PORT
    elif not isinstance(host, str):
        host = str(host)    # take care of the occasion where host is an ipaddress.IpAddress
    # pyro direct lookup
    port = config.NS_PORT if not port else port
    if URI.isUnixsockLocation(host):
        uristring = "PYRO:%s@%s" % (NAMESERVER_NAME, host)
    else:
        # if not a unix socket, check for ipv6
        if host and ":" in str(host):
            host = "[%s]" % host
        uristring = "PYRO:%s@%s:%d" % (NAMESERVER_NAME, host, port)
    uri = URI(uristring)
    log.debug("locating the NS: %s", uri)
    proxy = client.Proxy(uri)
    try:
        proxy._pyroBind()
        log.debug("located NS")
        return proxy
    except errors.PyroError as x:
        raise errors.NamingError("Failed to locate the nameserver") from x


def type_meta(class_or_object, prefix="class:"):
    """extracts type metadata from the given class or object, can be used as Name server metadata."""
    if hasattr(class_or_object, "__mro__"):
        return {prefix + c.__module__ + "." + c.__name__
                for c in class_or_object.__mro__ if c.__module__ not in ("builtins", "__builtin__")}
    if hasattr(class_or_object, "__class__"):
        return type_meta(class_or_object.__class__)
    return frozenset()



import time
from Pyro5.api import Proxy


with Proxy("PYRONAME:example.oneway2") as serv:
    print("incrementing a few times normally")
    serv.increment()
    serv.increment()
    serv.increment()
    counter = serv.getcount()
    print("counter in server is now: ", counter)
    print("incrementing a few times via oneway call (should be almost instantaneous)")
    serv.increment_oneway()
    serv.increment_oneway()
    serv.increment_oneway()
    counter2 = serv.getcount()
    print("counter is now: ", counter2)
    if counter2 == counter:
        print("ok, the oneway calls are still being processed in the background.")
        print("   we'll wait a bit till they're done...")
        time.sleep(2)
        counter2 = serv.getcount()
        print("counter is now: ", counter2)
    else:
        raise SystemExit("!? the oneway calls have not been processed in the background!?")



"""
The various serializers.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import array
import builtins
import uuid
import logging
import struct
import datetime
import decimal
import numbers
import inspect
import marshal
import json
import serpent
import contextlib
try:
    import msgpack
except ImportError:
    msgpack = None
from . import errors, config

__all__ = ["SerializerBase", "SerpentSerializer", "JsonSerializer", "MarshalSerializer", "MsgpackSerializer",
           "serializers", "serializers_by_id"]

log = logging.getLogger("Pyro5.serializers")


all_exceptions = {}
for name, t in vars(builtins).items():
    if type(t) is type and issubclass(t, BaseException):
        all_exceptions[name] = t
for name, t in vars(errors).items():
    if type(t) is type and issubclass(t, errors.PyroError):
        all_exceptions[name] = t


def pyro_class_serpent_serializer(obj, serializer, stream, level):
    # Override the default way that a Pyro URI/proxy/daemon is serialized.
    # Because it defines a __getstate__ it would otherwise just become a tuple,
    # and not be deserialized as a class.
    d = SerializerBase.class_to_dict(obj)
    serializer.ser_builtins_dict(d, stream, level)


def serialize_pyro_object_to_dict(obj):
    return {
        "__class__": "{:s}.{:s}".format(obj.__module__, obj.__class__.__name__),
        "state": obj.__getstate__()
    }


class SerializerBase(object):
    """Base class for (de)serializer implementations (which must be thread safe)"""
    serializer_id = 0  # define uniquely in subclass
    __custom_class_to_dict_registry = {}
    __custom_dict_to_class_registry = {}

    def loads(self, data):
        raise NotImplementedError("implement in subclass")

    def loadsCall(self, data):
        raise NotImplementedError("implement in subclass")

    def dumps(self, data):
        raise NotImplementedError("implement in subclass")

    def dumpsCall(self, obj, method, vargs, kwargs):
        raise NotImplementedError("implement in subclass")

    @classmethod
    def register_type_replacement(cls, object_type, replacement_function):
        raise NotImplementedError("implement in subclass")

    def _convertToBytes(self, data):
        if type(data) is bytearray:
            return bytes(data)
        if type(data) is memoryview:
            return data.tobytes()
        return data

    @classmethod
    def register_class_to_dict(cls, clazz, converter, serpent_too=True):
        """Registers a custom function that returns a dict representation of objects of the given class.
        The function is called with a single parameter; the object to be converted to a dict."""
        cls.__custom_class_to_dict_registry[clazz] = converter
        if serpent_too:
            with contextlib.suppress(errors.ProtocolError):
                def serpent_converter(obj, serializer, stream, level):
                    d = converter(obj)
                    serializer.ser_builtins_dict(d, stream, level)

                serpent.register_class(clazz, serpent_converter)

    @classmethod
    def unregister_class_to_dict(cls, clazz):
        """Removes the to-dict conversion function registered for the given class. Objects of the class
        will be serialized by the default mechanism again."""
        if clazz in cls.__custom_class_to_dict_registry:
            del cls.__custom_class_to_dict_registry[clazz]
        with contextlib.suppress(errors.ProtocolError):
            serpent.unregister_class(clazz)

    @classmethod
    def register_dict_to_class(cls, classname, converter):
        """
        Registers a custom converter function that creates objects from a dict with the given classname tag in it.
        The function is called with two parameters: the classname and the dictionary to convert to an instance of the class.
        """
        cls.__custom_dict_to_class_registry[classname] = converter

    @classmethod
    def unregister_dict_to_class(cls, classname):
        """
        Removes the converter registered for the given classname. Dicts with that classname tag
        will be deserialized by the default mechanism again.
        """
        if classname in cls.__custom_dict_to_class_registry:
            del cls.__custom_dict_to_class_registry[classname]

    @classmethod
    def class_to_dict(cls, obj):
        """
        Convert a non-serializable object to a dict. Partly borrowed from serpent.
        """
        for clazz in cls.__custom_class_to_dict_registry:
            if isinstance(obj, clazz):
                return cls.__custom_class_to_dict_registry[clazz](obj)
        if type(obj) in (set, dict, tuple, list):
            # we use a ValueError to mirror the exception type returned by serpent and other serializers
            raise ValueError("can't serialize type " + str(obj.__class__) + " into a dict")
        if hasattr(obj, "_pyroDaemon"):
            obj._pyroDaemon = None
        if isinstance(obj, BaseException):
            # special case for exceptions
            return {
                "__class__": obj.__class__.__module__ + "." + obj.__class__.__name__,
                "__exception__": True,
                "args": obj.args,
                "attributes": vars(obj)  # add custom exception attributes
            }
        # note: python 3.11+ object itself now has __getstate__
        has_own_getstate = (
            hasattr(type(obj), '__getstate__')
            and type(obj).__getstate__ is not getattr(object, '__getstate__', None)
        )
        if has_own_getstate:
            value = obj.__getstate__()
            if isinstance(value, dict):
                return value
        try:
            value = dict(vars(obj))  # make sure we can serialize anything that resembles a dict
            value["__class__"] = obj.__class__.__module__ + "." + obj.__class__.__name__
            return value
        except TypeError:
            if hasattr(obj, "__slots__"):
                # use the __slots__ instead of the vars dict
                value = {}
                for slot in obj.__slots__:
                    value[slot] = getattr(obj, slot)
                value["__class__"] = obj.__class__.__module__ + "." + obj.__class__.__name__
                return value
            else:
                raise errors.SerializeError("don't know how to serialize class " + str(obj.__class__) +
                                            " using serializer " + str(cls.__name__) +
                                            ". Give it vars() or an appropriate __getstate__")

    @classmethod
    def dict_to_class(cls, data):
        """
        Recreate an object out of a dict containing the class name and the attributes.
        Only a fixed set of classes are recognized.
        """
        from . import core, client, server  # circular imports...
        classname = data.get("__class__", "<unknown>")
        if isinstance(classname, bytes):
            classname = classname.decode("utf-8")
        if classname in cls.__custom_dict_to_class_registry:
            converter = cls.__custom_dict_to_class_registry[classname]
            return converter(classname, data)
        if "__" in classname:
            raise errors.SecurityError("refused to deserialize types with double underscores in their name: " + classname)
        # for performance reasons, the constructors below are hardcoded here
        # instead of added on a per-class basis to the dict-to-class registry
        if classname == "Pyro5.core.URI":
            uri = core.URI.__new__(core.URI)
            uri.__setstate__(data["state"])
            return uri
        elif classname == "Pyro5.client.Proxy":
            proxy = client.Proxy.__new__(client.Proxy)
            proxy.__setstate__(data["state"])
            return proxy
        elif classname == "Pyro5.server.Daemon":
            daemon = server.Daemon.__new__(server.Daemon)
            daemon.__setstate__(data["state"])
            return daemon
        elif classname.startswith("Pyro5.util."):
            if classname == "Pyro5.util.SerpentSerializer":
                return SerpentSerializer()
            elif classname == "Pyro5.util.MarshalSerializer":
                return MarshalSerializer()
            elif classname == "Pyro5.util.JsonSerializer":
                return JsonSerializer()
            elif classname == "Pyro5.util.MsgpackSerializer":
                return MsgpackSerializer()
        elif classname.startswith("Pyro5.errors."):
            errortype = getattr(errors, classname.split('.', 2)[2])
            if issubclass(errortype, errors.PyroError):
                return SerializerBase.make_exception(errortype, data)
        elif classname == "struct.error":
            return SerializerBase.make_exception(struct.error, data)
        elif classname == "Pyro5.core._ExceptionWrapper":
            ex = data["exception"]
            if isinstance(ex, dict) and "__class__" in ex:
                ex = SerializerBase.dict_to_class(ex)
            return core._ExceptionWrapper(ex)
        elif data.get("__exception__", False):
            if classname in all_exceptions:
                return SerializerBase.make_exception(all_exceptions[classname], data)
            # translate to the appropriate namespace...
            namespace, short_classname = classname.split('.', 1)
            if namespace in ("builtins", "exceptions"):
                exceptiontype = getattr(builtins, short_classname)
                if issubclass(exceptiontype, BaseException):
                    return SerializerBase.make_exception(exceptiontype, data)
            elif namespace == "sqlite3" and short_classname.endswith("Error"):
                import sqlite3
                exceptiontype = getattr(sqlite3, short_classname)
                if issubclass(exceptiontype, BaseException):
                    return SerializerBase.make_exception(exceptiontype, data)
        log.warning("unsupported serialized class: " + classname)
        raise errors.SerializeError("unsupported serialized class: " + classname)

    @staticmethod
    def make_exception(exceptiontype, data):
        ex = exceptiontype(*data["args"])
        if "attributes" in data:
            # restore custom attributes on the exception object
            for attr, value in data["attributes"].items():
                setattr(ex, attr, value)
        return ex

    def recreate_classes(self, literal):
        t = type(literal)
        if t is set:
            return {self.recreate_classes(x) for x in literal}
        if t is list:
            return [self.recreate_classes(x) for x in literal]
        if t is tuple:
            return tuple(self.recreate_classes(x) for x in literal)
        if t is dict:
            if "__class__" in literal:
                return self.dict_to_class(literal)
            result = {}
            for key, value in literal.items():
                result[key] = self.recreate_classes(value)
            return result
        return literal

    def __eq__(self, other):
        """this equality method is only to support the unit tests of this class"""
        return isinstance(other, SerializerBase) and vars(self) == vars(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    __hash__ = object.__hash__


class SerpentSerializer(SerializerBase):
    """(de)serializer that wraps the serpent serialization protocol."""
    serializer_id = 1  # never change this

    def dumpsCall(self, obj, method, vargs, kwargs):
        return serpent.dumps((obj, method, vargs, kwargs), module_in_classname=True, bytes_repr=config.SERPENT_BYTES_REPR)

    def dumps(self, data):
        return serpent.dumps(data, module_in_classname=True, bytes_repr=config.SERPENT_BYTES_REPR)

    def loadsCall(self, data):
        obj, method, vargs, kwargs = serpent.loads(data)
        vargs = self.recreate_classes(vargs)
        kwargs = self.recreate_classes(kwargs)
        return obj, method, vargs, kwargs

    def loads(self, data):
        return self.recreate_classes(serpent.loads(data))

    @classmethod
    def register_type_replacement(cls, object_type, replacement_function):
        def custom_serializer(obj, serpent_serializer, outputstream, indentlevel):
            replaced = replacement_function(obj)
            if replaced is obj:
                serpent_serializer.ser_default_class(replaced, outputstream, indentlevel)
            else:
                serpent_serializer._serialize(replaced, outputstream, indentlevel)

        if object_type is type or not inspect.isclass(object_type):
            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
        serpent.register_class(object_type, custom_serializer)

    @classmethod
    def dict_to_class(cls, data):
        if data.get("__class__") == "float":
            return float(data["value"])     # serpent encodes a float nan as a special class dict like this
        return super(SerpentSerializer, cls).dict_to_class(data)


class MarshalSerializer(SerializerBase):
    """(de)serializer that wraps the marshal serialization protocol."""
    serializer_id = 2  # never change this

    def dumpsCall(self, obj, method, vargs, kwargs):
        vargs = [self.convert_obj_into_marshallable(value) for value in vargs]
        kwargs = {key: self.convert_obj_into_marshallable(value) for key, value in kwargs.items()}
        return marshal.dumps((obj, method, vargs, kwargs))

    def dumps(self, data):
        return marshal.dumps(self.convert_obj_into_marshallable(data))

    def loadsCall(self, data):
        data = self._convertToBytes(data)
        obj, method, vargs, kwargs = marshal.loads(data)
        vargs = self.recreate_classes(vargs)
        kwargs = self.recreate_classes(kwargs)
        return obj, method, vargs, kwargs

    def loads(self, data):
        data = self._convertToBytes(data)
        return self.recreate_classes(marshal.loads(data))

    def convert_obj_into_marshallable(self, obj):
        marshalable_types = (str, int, float, type(None), bool, complex, bytes, bytearray,
                             tuple, set, frozenset, list, dict)
        if isinstance(obj, array.array):
            if obj.typecode == 'c':
                return obj.tostring()
            if obj.typecode == 'u':
                return obj.tounicode()
            return obj.tolist()
        if isinstance(obj, marshalable_types):
            return obj
        return self.class_to_dict(obj)

    @classmethod
    def class_to_dict(cls, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj)
        return super(MarshalSerializer, cls).class_to_dict(obj)

    @classmethod
    def register_type_replacement(cls, object_type, replacement_function):
        pass  # marshal serializer doesn't support per-type hooks


class JsonSerializer(SerializerBase):
    """(de)serializer that wraps the json serialization protocol."""
    serializer_id = 3  # never change this

    __type_replacements = {}

    def dumpsCall(self, obj, method, vargs, kwargs):
        data = {"object": obj, "method": method, "params": vargs, "kwargs": kwargs}
        data = json.dumps(data, ensure_ascii=False, default=self.default)
        return data.encode("utf-8")

    def dumps(self, data):
        data = json.dumps(data, ensure_ascii=False, default=self.default)
        return data.encode("utf-8")

    def loadsCall(self, data):
        data = self._convertToBytes(data).decode("utf-8")
        data = json.loads(data)
        vargs = self.recreate_classes(data["params"])
        kwargs = self.recreate_classes(data["kwargs"])
        return data["object"], data["method"], vargs, kwargs

    def loads(self, data):
        data = self._convertToBytes(data).decode("utf-8")
        return self.recreate_classes(json.loads(data))

    def default(self, obj):
        replacer = self.__type_replacements.get(type(obj), None)
        if replacer:
            obj = replacer(obj)
        if isinstance(obj, set):
            return tuple(obj)  # json module can't deal with sets so we make a tuple out of it
        if isinstance(obj, uuid.UUID):
            return str(obj)
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        if isinstance(obj, decimal.Decimal):
            return str(obj)
        if isinstance(obj, array.array):
            if obj.typecode == 'c':
                return obj.tostring()
            if obj.typecode == 'u':
                return obj.tounicode()
            return obj.tolist()
        return self.class_to_dict(obj)

    @classmethod
    def register_type_replacement(cls, object_type, replacement_function):
        if object_type is type or not inspect.isclass(object_type):
            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
        cls.__type_replacements[object_type] = replacement_function


class MsgpackSerializer(SerializerBase):
    """(de)serializer that wraps the msgpack serialization protocol."""
    serializer_id = 4  # never change this

    __type_replacements = {}

    def dumpsCall(self, obj, method, vargs, kwargs):
        return msgpack.packb((obj, method, vargs, kwargs), use_bin_type=True, default=self.default)

    def dumps(self, data):
        return msgpack.packb(data, use_bin_type=True, default=self.default)

    def loadsCall(self, data):
        return msgpack.unpackb(self._convertToBytes(data), raw=False, object_hook=self.object_hook)

    def loads(self, data):
        return msgpack.unpackb(self._convertToBytes(data), raw=False, object_hook=self.object_hook, ext_hook=self.ext_hook)

    def default(self, obj):
        replacer = self.__type_replacements.get(type(obj), None)
        if replacer:
            obj = replacer(obj)
        if isinstance(obj, set):
            return tuple(obj)  # msgpack module can't deal with sets so we make a tuple out of it
        if isinstance(obj, uuid.UUID):
            return str(obj)
        if isinstance(obj, complex):
            return msgpack.ExtType(0x30, struct.pack("dd", obj.real, obj.imag))
        if isinstance(obj, datetime.datetime):
            if obj.tzinfo:
                raise errors.SerializeError("msgpack cannot serialize datetime with timezone info")
            return msgpack.ExtType(0x32, struct.pack("d", obj.timestamp()))
        if isinstance(obj, datetime.date):
            return msgpack.ExtType(0x33, struct.pack("l", obj.toordinal()))
        if isinstance(obj, decimal.Decimal):
            return str(obj)
        if isinstance(obj, numbers.Number):
            return msgpack.ExtType(0x31, str(obj).encode("ascii"))     # long
        if isinstance(obj, array.array):
            if obj.typecode == 'c':
                return obj.tostring()
            if obj.typecode == 'u':
                return obj.tounicode()
            return obj.tolist()
        return self.class_to_dict(obj)

    def object_hook(self, obj):
        if "__class__" in obj:
            return self.dict_to_class(obj)
        return obj

    def ext_hook(self, code, data):
        if code == 0x30:
            real, imag = struct.unpack("dd", data)
            return complex(real, imag)
        if code == 0x31:
            return int(data)
        if code == 0x32:
            return datetime.datetime.fromtimestamp(struct.unpack("d", data)[0])
        if code == 0x33:
            return datetime.date.fromordinal(struct.unpack("l", data)[0])
        raise errors.SerializeError("invalid ext code for msgpack: " + str(code))

    @classmethod
    def register_type_replacement(cls, object_type, replacement_function):
        if object_type is type or not inspect.isclass(object_type):
            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
        cls.__type_replacements[object_type] = replacement_function


"""The various serializers that are supported"""
serializers = {
    "serpent": SerpentSerializer(),
    "marshal": MarshalSerializer(),
    "json": JsonSerializer()
}

if msgpack:
    serializers["msgpack"] = MsgpackSerializer()


"""The available serializers by their internal id"""
serializers_by_id = {ser.serializer_id: ser for ser in serializers.values()}



# This is the code that visits the warehouse.
import sys
import Pyro5.errors
from Pyro5.api import Proxy
from person import Person


sys.excepthook = Pyro5.errors.excepthook

warehouse = Proxy("PYRONAME:example.warehouse")
janet = Person("Janet")
henry = Person("Henry")
janet.visit(warehouse)
henry.visit(warehouse)



"""
Pyro package. Some generic init stuff to set up logging etc.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

__version__ = "5.16"
__author__ = "Irmen de Jong"


def __configure_logging():
    """Do some basic config of the logging module at package import time.
    The configuring is done only if the PYRO_LOGLEVEL env var is set.
    If you want to use your own logging config, make sure you do
    that before any Pyro imports. Then Pyro will skip the autoconfig.
    Set the env var PYRO_LOGFILE to change the name of the autoconfigured
    log file (default is pyro5.log in the current dir). Use '{stderr}' to
    make the log go to the standard error output."""
    import os
    import logging

    level = os.environ.get("PYRO_LOGLEVEL")
    logfilename = os.environ.get("PYRO_LOGFILE", "pyro5.log")
    if level:
        levelvalue = getattr(logging, level)
        if len(logging.root.handlers) == 0:
            logging.basicConfig(
                level=levelvalue,
                filename=None if logfilename == "{stderr}" else logfilename,
                datefmt="%Y-%m-%d %H:%M:%S",
                format="[%(asctime)s.%(msecs)03d,%(name)s,%(levelname)s] %(message)s"
            )
            log = logging.getLogger("Pyro5")
            log.info("Pyro log configured using built-in defaults, level=%s", level)
    else:
        # PYRO_LOGLEVEL is not set, disable Pyro logging. No message is printed about this fact.
        log = logging.getLogger("Pyro5")
        log.setLevel(9999)
        return logfilename, None
    return logfilename, level or None


_pyro_logfile, _pyro_loglevel = __configure_logging()

from .configure import global_config as config



import time
from Pyro5.api import Proxy


with Proxy("PYRONAME:example.oneway") as serv:
    print("starting server using a oneway call")
    serv.oneway_start(6)
    print("doing some more oneway calls inbetween (this should be finished really quick)")
    serv.nothing()
    serv.nothing()
    serv.nothing()
    serv.nothing()
    print("oneway calls done, this should have taken almost no time.")

    time.sleep(2)
    print("\nNow contacting the server to see if it's done.")
    print("we are faster, so you should see a few attempts,")
    print("until the server is finished.")
    while True:
        print("server done?")
        if serv.ready():
            print("yes!")
            break
        else:
            print("no, trying again")
            time.sleep(1)

    print("getting the result from the server: %s" % serv.result())
    print("\nCalling oneway work method, server will continue working while we are done "
          "(quickly check the server console output now to see it running!).")
    serv.oneway_work()



"""
Tests for the thread pool.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import time
import random
import pytest
from Pyro5 import socketutil, server
from Pyro5.svr_threads import Pool, PoolError, NoFreeWorkersError, SocketServer_Threadpool
from Pyro5 import config


JOB_TIME = 0.2


class Job(object):
    def __init__(self, name="unnamed"):
        self.name = name

    def __call__(self):
        time.sleep(JOB_TIME - random.random() / 10.0)


class SlowJob(object):
    def __init__(self, name="unnamed"):
        self.name = name

    def __call__(self):
        time.sleep(5*JOB_TIME - random.random() / 10.0)


class TestThreadPool:
    def setup_method(self):
        config.THREADPOOL_SIZE_MIN = 2
        config.THREADPOOL_SIZE = 4

    def teardown_method(self):
        config.reset()

    def testCreate(self):
        with Pool() as jq:
            _ = repr(jq)
        assert jq.closed

    def testSingle(self):
        with Pool() as p:
            job = Job()
            p.process(job)
            time.sleep(0.02)  # let it pick up the job
            assert len(p.busy) == 1

    def testAllBusy(self):
        try:
            config.COMMTIMEOUT = 0.2
            with Pool() as p:
                for i in range(config.THREADPOOL_SIZE):
                    p.process(SlowJob(str(i+1)))
                # putting one more than the number of workers should raise an error:
                with pytest.raises(NoFreeWorkersError):
                    p.process(SlowJob("toomuch"))
        finally:
            config.COMMTIMEOUT = 0.0

    def testClose(self):
        with Pool() as p:
            for i in range(config.THREADPOOL_SIZE):
                p.process(Job(str(i + 1)))
        with pytest.raises(PoolError):
            p.process(Job("1"))  # must not allow new jobs after closing
        assert len(p.busy) == 0
        assert len(p.idle) == 0

    def testScaling(self):
        with Pool() as p:
            for i in range(config.THREADPOOL_SIZE_MIN-1):
                p.process(Job("x"))
            assert len(p.idle) == 1
            assert len(p.busy) == config.THREADPOOL_SIZE_MIN-1
            p.process(Job("x"))
            assert len(p.idle) == 0
            assert len(p.busy) == config.THREADPOOL_SIZE_MIN
            # grow until no more free workers
            while True:
                try:
                    p.process(Job("x"))
                except NoFreeWorkersError:
                    break
            assert len(p.idle) == 0
            assert len(p.busy) == config.THREADPOOL_SIZE
            # wait till jobs are done and check ending situation
            time.sleep(JOB_TIME*1.5)
            assert len(p.busy) == 0
            assert len(p.idle) == config.THREADPOOL_SIZE_MIN


class ServerCallback(server.Daemon):
    def __init__(self):
        super().__init__()
        self.received_denied_reasons = []

    def _handshake(self, connection, denied_reason=None):
        self.received_denied_reasons.append(denied_reason)  # store the denied reason
        return True

    def handleRequest(self, connection):
        time.sleep(0.05)

    def _housekeeping(self):
        pass


class TestThreadPoolServer:
    def setup_method(self):
        config.THREADPOOL_SIZE_MIN = 1
        config.THREADPOOL_SIZE = 1
        config.POLLTIMEOUT = 0.5
        config.COMMTIMEOUT = 0.5

    def teardown_method(self):
        config.reset()

    def testServerPoolFull(self):
        port = socketutil.find_probably_unused_port()
        serv = SocketServer_Threadpool()
        daemon = ServerCallback()
        serv.init(daemon, "localhost", port)
        serversock = serv.sock.getsockname()
        csock1 = socketutil.create_socket(connect=serversock)
        csock2 = socketutil.create_socket(connect=serversock)
        try:
            serv.events([serv.sock])
            time.sleep(0.2)
            assert daemon.received_denied_reasons == [None]
            serv.events([serv.sock])
            time.sleep(0.2)
            assert len(daemon.received_denied_reasons) == 2
            assert "no free workers, increase server threadpool size" in daemon.received_denied_reasons
        finally:
            csock1.close()
            csock2.close()
            serv.shutdown()




import Pyro5.api


uri = input("enter factory server object uri: ").strip()
factory = Pyro5.api.Proxy(uri)

# create several things.
print("Creating things.")
thing1 = factory.createSomething(1)
thing2 = factory.createSomething(2)
thing3 = factory.createSomething(3)

print(repr(thing1))

# interact with them on the server.
print("Speaking stuff (see output on server).")
thing1.speak("I am the first")
thing2.speak("I am second")
thing3.speak("I am last then...")



"""
Single module that centralizes the main symbols from the Pyro5 API.
It imports most of the other packages that it needs
and provides shortcuts to the most frequently used objects and functions from those packages.
This means you can mostly just ``import Pyro5.api`` in your code to have access to most of
the Pyro5 objects and functions.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

from . import __version__
from .configure import global_config as config
from .core import URI, locate_ns, resolve, type_meta
from .client import Proxy, BatchProxy, SerializedBlob
from .server import Daemon, DaemonObject, callback, expose, behavior, oneway, serve
from .nameserver import start_ns, start_ns_loop
from .serializers import SerializerBase
from .callcontext import current_context

register_dict_to_class = SerializerBase.register_dict_to_class
register_class_to_dict = SerializerBase.register_class_to_dict
unregister_dict_to_class = SerializerBase.unregister_dict_to_class
unregister_class_to_dict = SerializerBase.unregister_class_to_dict


__all__ = ["config", "URI", "locate_ns", "resolve", "type_meta", "current_context",
           "Proxy", "BatchProxy", "SerializedBlob", "SerializerBase",
           "Daemon", "DaemonObject", "callback", "expose", "behavior", "oneway",
           "start_ns", "start_ns_loop", "serve", "register_dict_to_class",
           "register_class_to_dict", "unregister_dict_to_class", "unregister_class_to_dict"]



"""
Server related classes (Daemon etc)

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import os
import sys
import uuid
import time
import socket
import collections
import threading
import logging
import inspect
import warnings
import weakref
import serpent
import ipaddress
from typing import TypeVar, Tuple, Union, Optional, Dict, Any, Sequence, Set
from . import config, core, errors, serializers, socketutil, protocol, client
from .callcontext import current_context
from collections.abc import Callable

__all__ = ["Daemon", "DaemonObject", "callback", "expose", "behavior", "oneway", "serve"]

log = logging.getLogger("Pyro5.server")


_private_dunder_methods = frozenset([
    "__init__", "__init_subclass__", "__class__", "__module__", "__weakref__",
    "__call__", "__new__", "__del__", "__repr__",
    "__str__", "__format__", "__nonzero__", "__bool__", "__coerce__",
    "__cmp__", "__eq__", "__ne__", "__hash__", "__ge__", "__gt__", "__le__", "__lt__",
    "__dir__", "__enter__", "__exit__", "__copy__", "__deepcopy__", "__sizeof__",
    "__getattr__", "__setattr__", "__hasattr__", "__getattribute__", "__delattr__",
    "__instancecheck__", "__subclasscheck__", "__getinitargs__", "__getnewargs__",
    "__getstate__", "__setstate__", "__reduce__", "__reduce_ex__", "__subclasshook__"
])


def is_private_attribute(attr_name: str) -> bool:
    """returns if the attribute name is to be considered private or not."""
    if attr_name in _private_dunder_methods:
        return True
    if not attr_name.startswith('_'):
        return False
    if len(attr_name) > 4 and attr_name.startswith("__") and attr_name.endswith("__"):
        return False
    return True


# decorators

def callback(method: Callable) -> Callable:
    """
    decorator to mark a method to be a 'callback'. This will make Pyro
    raise any errors also on the callback side, and not only on the side
    that does the callback call.
    """
    method._pyroCallback = True     # type: ignore
    return method


def oneway(method: Callable) -> Callable:
    """
    decorator to mark a method to be oneway (client won't wait for a response)
    """
    method._pyroOneway = True       # type: ignore
    return method


_T = TypeVar("_T", bound=Union[Callable, type])


def expose(method_or_class: _T) -> _T:
    """
    Decorator to mark a method or class to be exposed for remote calls.
    You can apply it to a method or a class as a whole.
    If you need to change the default instance mode or instance creator, also use a @behavior decorator.
    """
    if inspect.isdatadescriptor(method_or_class):
        func = method_or_class.fget or method_or_class.fset or method_or_class.fdel     # type: ignore
        if is_private_attribute(func.__name__):
            raise AttributeError("exposing private names (starting with _) is not allowed")
        func._pyroExposed = True
        return method_or_class
    attrname = getattr(method_or_class, "__name__", None)
    if not attrname or isinstance(method_or_class, (classmethod, staticmethod)):
        # we could be dealing with a descriptor (classmethod/staticmethod), this means the order of the decorators is wrong
        if inspect.ismethoddescriptor(method_or_class):
            attrname = method_or_class.__get__(None, dict).__name__     # type: ignore
            raise AttributeError("using @expose on a classmethod/staticmethod must be done "
                                 "after @classmethod/@staticmethod. Method: " + attrname)
        else:
            raise AttributeError("@expose cannot determine what this is: "+repr(method_or_class))
    if is_private_attribute(attrname):
        raise AttributeError("exposing private names (starting with _) is not allowed")
    if inspect.isclass(method_or_class):
        clazz = method_or_class
        log.debug("exposing all members of %r", clazz)
        for name in clazz.__dict__:
            if is_private_attribute(name):
                continue
            thing = getattr(clazz, name)
            if inspect.isfunction(thing) or inspect.ismethoddescriptor(thing):
                thing._pyroExposed = True
            elif inspect.ismethod(thing):
                thing.__func__._pyroExposed = True
            elif inspect.isdatadescriptor(thing):
                if getattr(thing, "fset", None):
                    thing.fset._pyroExposed = True
                if getattr(thing, "fget", None):
                    thing.fget._pyroExposed = True
                if getattr(thing, "fdel", None):
                    thing.fdel._pyroExposed = True
        clazz._pyroExposed = True       # type: ignore
        return clazz
    method_or_class._pyroExposed = True     # type: ignore
    return method_or_class


def behavior(instance_mode: str = "session", instance_creator: Optional[Callable] = None) -> Callable:
    """
    Decorator to specify the server behavior of your Pyro class.
    """
    def _behavior(clazz):
        if not inspect.isclass(clazz):
            raise TypeError("behavior decorator can only be used on a class")
        if instance_mode not in ("single", "session", "percall"):
            raise ValueError("invalid instance mode: " + instance_mode)
        if instance_creator and not callable(instance_creator):
            raise TypeError("instance_creator must be a callable")
        clazz._pyroInstancing = (instance_mode, instance_creator)
        return clazz
    if not isinstance(instance_mode, str):
        raise SyntaxError("behavior decorator is missing argument(s)")
    return _behavior


@expose
class DaemonObject(object):
    """The part of the daemon that is exposed as a Pyro object."""

    def __init__(self, daemon):
        self.daemon = daemon

    def registered(self):
        """returns a list of all object names registered in this daemon"""
        return list(self.daemon.objectsById.keys())

    def ping(self):
        """a simple do-nothing method for testing purposes"""
        pass

    def info(self):
        """return some descriptive information about the daemon"""
        return "%s bound on %s, NAT %s, %d objects registered. Servertype: %s" % (
            core.DAEMON_NAME, self.daemon.locationStr, self.daemon.natLocationStr,
            len(self.daemon.objectsById), self.daemon.transportServer)

    def get_metadata(self, objectId):
        """
        Get metadata for the given object (exposed methods, oneways, attributes).
        """
        obj = _unpack_weakref(self.daemon.objectsById.get(objectId))
        if obj is not None:
            metadata = _get_exposed_members(obj)
            if not metadata["methods"] and not metadata["attrs"]:
                # Something seems wrong: nothing is remotely exposed.
                warnings.warn("Class %s doesn't expose any methods or attributes. Did you forget setting @expose on them?" % repr(obj))
            return metadata
        else:
            log.debug("unknown object requested: %s", objectId)
            raise errors.DaemonError("unknown object")

    def get_next_stream_item(self, streamId):
        if streamId not in self.daemon.streaming_responses:
            raise errors.PyroError("item stream terminated")
        client, timestamp, linger_timestamp, stream = self.daemon.streaming_responses[streamId]
        if client is None:
            # reset client connection association (can be None if proxy disconnected)
            self.daemon.streaming_responses[streamId] = (current_context.client, timestamp, 0, stream)
        try:
            return next(stream)
        except Exception:
            # in case of error (or StopIteration!) the stream is removed
            del self.daemon.streaming_responses[streamId]
            raise

    def close_stream(self, streamId):
        if streamId in self.daemon.streaming_responses:
            del self.daemon.streaming_responses[streamId]


class Daemon(object):
    """
    Pyro daemon. Contains server side logic and dispatches incoming remote method calls
    to the appropriate objects.
    """

    def __init__(self, host=None, port=0, unixsocket=None, nathost=None, natport=None, interface=DaemonObject, connected_socket=None):
        if connected_socket:
            nathost = natport = None
        else:
            if host is None:
                host = config.HOST
            elif not isinstance(host, str):
                host = str(host)  # take care of the occasion where host is an ipaddress.IpAddress
            if nathost is None:
                nathost = config.NATHOST
            elif not isinstance(nathost, str):
                nathost = str(nathost)  # take care of the occasion where host is an ipaddress.IpAddress
            if natport is None and nathost is not None:
                natport = config.NATPORT
            if nathost and unixsocket:
                raise ValueError("cannot use nathost together with unixsocket")
            if (nathost is None) ^ (natport is None):
                raise ValueError("must provide natport with nathost")
        self.__mustshutdown = threading.Event()
        self.__mustshutdown.set()
        self.__loopstopped = threading.Event()
        self.__loopstopped.set()
        if connected_socket:
            from .svr_existingconn import SocketServer_ExistingConnection
            self.transportServer = SocketServer_ExistingConnection()
            self.transportServer.init(self, connected_socket)
        else:
            if config.SERVERTYPE == "thread":
                from .svr_threads import SocketServer_Threadpool
                self.transportServer = SocketServer_Threadpool()
            elif config.SERVERTYPE == "multiplex":
                from .svr_multiplex import SocketServer_Multiplex
                self.transportServer = SocketServer_Multiplex()
            else:
                raise errors.PyroError("invalid server type '%s'" % config.SERVERTYPE)
            self.transportServer.init(self, host, port, unixsocket)
        #: The location (str of the form ``host:portnumber``) on which the Daemon is listening
        self.locationStr = self.transportServer.locationStr
        log.debug("daemon created on %s - %s (pid %d)", self.locationStr, socketutil.family_str(self.transportServer.sock), os.getpid())
        natport_for_loc = natport
        if natport == 0:
            # expose internal port number as NAT port as well. (don't use port because it could be 0 and will be chosen by the OS)
            natport_for_loc = int(self.locationStr.split(":")[1])
        # The NAT-location (str of the form ``nathost:natportnumber``) on which the Daemon is exposed for use with NAT-routing
        self.natLocationStr = "%s:%d" % (nathost, natport_for_loc) if nathost else None
        if self.natLocationStr:
            log.debug("NAT address is %s", self.natLocationStr)
        pyroObject = interface(self)
        pyroObject._pyroId = core.DAEMON_NAME
        # Dictionary from Pyro object id to the actual Pyro object registered by this id
        self.objectsById = {pyroObject._pyroId: pyroObject}
        log.debug("pyro protocol version: %d" % protocol.PROTOCOL_VERSION)
        self._pyroInstances = {}   # pyro objects for instance_mode=single (singletons, just one per daemon)
        self.streaming_responses = {}   # stream_id -> (client, creation_timestamp, linger_timestamp, stream)
        self.housekeeper_lock = threading.Lock()
        self.create_single_instance_lock = threading.Lock()
        self.__mustshutdown.clear()
        self.methodcall_error_handler = _default_methodcall_error_handler

    @property
    def sock(self):
        """the server socket used by the daemon"""
        return self.transportServer.sock

    @property
    def sockets(self):
        """list of all sockets used by the daemon (server socket and all active client sockets)"""
        return self.transportServer.sockets

    @property
    def selector(self):
        """the multiplexing selector used, if using the multiplex server type"""
        return self.transportServer.selector

    @staticmethod
    def serveSimple(objects, host=None, port=0, daemon=None, ns=True, verbose=True) -> None:
        """
        Backwards compatibility method to fire up a daemon and start serving requests.
        New code should just use the global ``serve`` function instead.
        """
        serve(objects, host, port, daemon, ns, verbose)

    def requestLoop(self, loopCondition=lambda: True) -> None:
        """
        Goes in a loop to service incoming requests, until someone breaks this
        or calls shutdown from another thread.
        """
        self.__mustshutdown.clear()
        log.info("daemon %s entering requestloop", self.locationStr)
        try:
            self.__loopstopped.clear()
            self.transportServer.loop(loopCondition=lambda: not self.__mustshutdown.is_set() and loopCondition())
        finally:
            self.__loopstopped.set()
        log.debug("daemon exits requestloop")

    def events(self, eventsockets):
        """for use in an external event loop: handle any requests that are pending for this daemon"""
        return self.transportServer.events(eventsockets)

    def shutdown(self):
        """Cleanly terminate a daemon that is running in the requestloop."""
        log.debug("daemon shutting down")
        self.streaming_responses = {}
        time.sleep(0.02)
        self.__mustshutdown.set()
        if self.transportServer:
            self.transportServer.shutdown()
            time.sleep(0.02)
        self.close()
        self.__loopstopped.wait(timeout=5)  # use timeout to avoid deadlock situations

    @property
    def _shutting_down(self):
        return self.__mustshutdown.is_set()

    def _handshake(self, conn, denied_reason=None):
        """
        Perform connection handshake with new clients.
        Client sends a MSG_CONNECT message with a serialized data payload.
        If all is well, return with a CONNECT_OK message.
        The reason we're not doing this with a MSG_INVOKE method call on the daemon
        (like when retrieving the metadata) is because we need to force the clients
        to get past an initial connect handshake before letting them invoke any method.
        Return True for successful handshake, False if something was wrong.
        If a denied_reason is given, the handshake will fail with the given reason.
        """
        serializer_id = serializers.MarshalSerializer.serializer_id
        msg_seq = 0
        try:
            msg = protocol.recv_stub(conn, [protocol.MSG_CONNECT])
            msg_seq = msg.seq
            if denied_reason:
                raise Exception(denied_reason)
            if config.LOGWIRE:
                protocol.log_wiredata(log, "daemon handshake received", msg)
            if msg.flags & protocol.FLAGS_CORR_ID:
                current_context.correlation_id = uuid.UUID(bytes=msg.corr_id)
            else:
                current_context.correlation_id = uuid.uuid4()
            serializer_id = msg.serializer_id
            serializer = serializers.serializers_by_id[serializer_id]
            data = serializer.loads(msg.data)
            handshake_response = self.validateHandshake(conn, data["handshake"])
            handshake_response = {
                "handshake": handshake_response,
                "meta": self.objectsById[core.DAEMON_NAME].get_metadata(data["object"])
            }
            data = serializer.dumps(handshake_response)
            msgtype = protocol.MSG_CONNECTOK
        except errors.ConnectionClosedError:
            log.debug("handshake failed, connection closed early")
            return False
        except Exception as x:
            log.debug("handshake failed, reason:", exc_info=True)
            serializer = serializers.serializers_by_id[serializer_id]
            data = serializer.dumps(str(x))
            msgtype = protocol.MSG_CONNECTFAIL
        # We need a minimal amount of response data or the socket will remain blocked
        # on some systems... (messages smaller than 40 bytes)
        msg = protocol.SendingMessage(msgtype, 0, msg_seq, serializer_id, data, annotations=self.__annotations())
        if config.LOGWIRE:
            protocol.log_wiredata(log, "daemon handshake response", msg)
        conn.send(msg.data)
        return msg.type == protocol.MSG_CONNECTOK

    def validateHandshake(self, conn, data):
        """
        Override this to create a connection validator for new client connections.
        It should return a response data object normally if the connection is okay,
        or should raise an exception if the connection should be denied.
        """
        return "hello"

    def clientDisconnect(self, conn):
        """
        Override this to handle a client disconnect.
        Conn is the SocketConnection object that was disconnected.
        """
        pass

    def handleRequest(self, conn):
        """
        Handle incoming Pyro request. Catches any exception that may occur and
        wraps it in a reply to the calling side, as to not make this server side loop
        terminate due to exceptions caused by remote invocations.
        """
        request_flags = 0
        request_seq = 0
        request_serializer_id = serializers.MarshalSerializer.serializer_id
        wasBatched = False
        isCallback = False
        try:
            msg = protocol.recv_stub(conn, [protocol.MSG_INVOKE, protocol.MSG_PING])
        except errors.CommunicationError as x:
            # we couldn't even get data from the client, this is an immediate error
            # log.info("error receiving data from client %s: %s", conn.sock.getpeername(), x)
            raise x
        try:
            request_flags = msg.flags
            request_seq = msg.seq
            request_serializer_id = msg.serializer_id
            if msg.flags & protocol.FLAGS_CORR_ID:
                current_context.correlation_id = uuid.UUID(bytes=msg.corr_id)
            else:
                current_context.correlation_id = uuid.uuid4()
            if config.LOGWIRE:
                protocol.log_wiredata(log, "daemon wiredata received", msg)
            if msg.type == protocol.MSG_PING:
                # return same seq, but ignore any data (it's a ping, not an echo). Nothing is deserialized.
                msg = protocol.SendingMessage(protocol.MSG_PING, 0, msg.seq, msg.serializer_id, b"pong", annotations=self.__annotations())
                if config.LOGWIRE:
                    protocol.log_wiredata(log, "daemon wiredata sending", msg)
                conn.send(msg.data)
                return
            serializer = serializers.serializers_by_id[msg.serializer_id]
            if request_flags & protocol.FLAGS_KEEPSERIALIZED:
                # pass on the wire protocol message blob unchanged
                objId, method, vargs, kwargs = self.__deserializeBlobArgs(msg)
            else:
                # normal deserialization of remote call arguments
                objId, method, vargs, kwargs = serializer.loadsCall(msg.data)
            current_context.client = conn
            try:
                # store, because on oneway calls, socket will be disconnected:
                current_context.client_sock_addr = conn.sock.getpeername()
            except socket.error:
                current_context.client_sock_addr = None  # sometimes getpeername() doesn't work...
            current_context.seq = msg.seq
            current_context.annotations = msg.annotations
            current_context.msg_flags = msg.flags
            current_context.serializer_id = msg.serializer_id
            del msg  # invite GC to collect the object, don't wait for out-of-scope
            obj = _unpack_weakref(self.objectsById.get(objId))
            if obj is not None:
                if inspect.isclass(obj):
                    obj = self._getInstance(obj, conn)
                if request_flags & protocol.FLAGS_BATCH:
                    # batched method calls, loop over them all and collect all results
                    data = []
                    for method, vargs, kwargs in vargs:
                        method = _get_attribute(obj, method)
                        try:
                            result = method(*vargs, **kwargs)  # this is the actual method call to the Pyro object
                        except Exception as xv:
                            self.methodcall_error_handler(self, current_context.client_sock_addr, method, vargs, kwargs, xv)
                            xv._pyroTraceback = errors.format_traceback(detailed=config.DETAILED_TRACEBACK)
                            data.append(core._ExceptionWrapper(xv))
                            break  # stop processing the rest of the batch
                        else:
                            data.append(result)    # note that we don't support streaming results in batch mode
                    wasBatched = True
                else:
                    # normal single method call
                    if method == "__getattr__":
                        # special case for direct attribute access (only exposed @properties are accessible)
                        data = _get_exposed_property_value(obj, vargs[0])
                        if not request_flags & protocol.FLAGS_ONEWAY:
                            isStream, data = self._streamResponse(data, conn)
                            if isStream:
                                # throw an exception as well as setting message flags
                                # this way, it is backwards compatible with older pyro versions.
                                exc = errors.ProtocolError("result of call is an iterator")
                                ann = {"STRM": data.encode()} if data else {}
                                self._sendExceptionResponse(conn, request_seq, serializer.serializer_id, exc, None,
                                                            annotations=ann, flags=protocol.FLAGS_ITEMSTREAMRESULT)
                                return
                    elif method == "__setattr__":
                        # special case for direct attribute access (only exposed @properties are accessible)
                        data = _set_exposed_property_value(obj, vargs[0], vargs[1])
                    else:
                        method = _get_attribute(obj, method)
                        if request_flags & protocol.FLAGS_ONEWAY:
                            # oneway call to be run inside its own thread, otherwise client blocking can still occur
                            #    on the next call on the same proxy
                            _OnewayCallThread(method, vargs, kwargs, self, current_context.client_sock_addr).start()
                        else:
                            isCallback = getattr(method, "_pyroCallback", False)
                            try:
                                data = method(*vargs, **kwargs)  # this is the actual method call to the Pyro object
                            except Exception as xv:
                                self.methodcall_error_handler(self, current_context.client_sock_addr, method, vargs, kwargs, xv)
                                raise
                            if not request_flags & protocol.FLAGS_ONEWAY:
                                isStream, data = self._streamResponse(data, conn)
                                if isStream:
                                    # throw an exception as well as setting message flags
                                    # this way, it is backwards compatible with older pyro versions.
                                    exc = errors.ProtocolError("result of call is an iterator")
                                    ann = {"STRM": data.encode()} if data else {}
                                    self._sendExceptionResponse(conn, request_seq, serializer.serializer_id, exc, None,
                                                                annotations=ann, flags=protocol.FLAGS_ITEMSTREAMRESULT)
                                    return
            else:
                log.debug("unknown object requested: %s", objId)
                raise errors.DaemonError("unknown object")
            if request_flags & protocol.FLAGS_ONEWAY:
                return  # oneway call, don't send a response
            else:
                data = serializer.dumps(data)
                response_flags = 0
                if wasBatched:
                    response_flags |= protocol.FLAGS_BATCH
                msg = protocol.SendingMessage(protocol.MSG_RESULT, response_flags, request_seq, serializer.serializer_id, data,
                                              annotations=self.__annotations())
                current_context.response_annotations = {}
                if config.LOGWIRE:
                    protocol.log_wiredata(log, "daemon wiredata sending", msg)
                conn.send(msg.data)
        except Exception as xv:
            msg = getattr(xv, "pyroMsg", None)
            if msg:
                request_seq = msg.seq
                request_serializer_id = msg.serializer_id
            if not isinstance(xv, errors.ConnectionClosedError):
                if not request_flags & protocol.FLAGS_ONEWAY:
                    if isinstance(xv, errors.SerializeError) or not isinstance(xv, errors.CommunicationError):
                        # only return the error to the client if it wasn't a oneway call, and not a communication error
                        # (in these cases, it makes no sense to try to report the error back to the client...)
                        tblines = errors.format_traceback(detailed=config.DETAILED_TRACEBACK)
                        self._sendExceptionResponse(conn, request_seq, request_serializer_id, xv, tblines)
            if isCallback or isinstance(xv, (errors.CommunicationError, errors.SecurityError)):
                raise  # re-raise if flagged as callback, communication or security error.

    def _clientDisconnect(self, conn):
        if config.ITER_STREAM_LINGER > 0:
            # client goes away, keep streams around for a bit longer (allow reconnect)
            for streamId in list(self.streaming_responses):
                info = self.streaming_responses.get(streamId, None)
                if info and info[0] is conn:
                    _, timestamp, _, stream = info
                    self.streaming_responses[streamId] = (None, timestamp, time.time(), stream)
        else:
            # client goes away, close any streams it had open as well
            for streamId in list(self.streaming_responses):
                info = self.streaming_responses.get(streamId, None)
                if info and info[0] is conn:
                    del self.streaming_responses[streamId]
        self.clientDisconnect(conn)  # user overridable hook

    def _housekeeping(self):
        """
        Perform periodical housekeeping actions (cleanups etc)
        """
        if self._shutting_down:
            return
        with self.housekeeper_lock:
            if self.streaming_responses:
                if config.ITER_STREAM_LIFETIME > 0:
                    # cleanup iter streams that are past their lifetime
                    for streamId in list(self.streaming_responses.keys()):
                        info = self.streaming_responses.get(streamId, None)
                        if info:
                            last_use_period = time.time() - info[1]
                            if 0 < config.ITER_STREAM_LIFETIME < last_use_period:
                                del self.streaming_responses[streamId]
                if config.ITER_STREAM_LINGER > 0:
                    # cleanup iter streams that are past their linger time
                    for streamId in list(self.streaming_responses.keys()):
                        info = self.streaming_responses.get(streamId, None)
                        if info and info[2]:
                            linger_period = time.time() - info[2]
                            if linger_period > config.ITER_STREAM_LINGER:
                                del self.streaming_responses[streamId]
            self.housekeeping()

    def housekeeping(self):
        """
        Override this to add custom periodic housekeeping (cleanup) logic.
        This will be called every few seconds by the running daemon's request loop.
        """
        pass

    def _getInstance(self, clazz, conn):
        """
        Find or create a new instance of the class
        """
        def createInstance(clazz, creator):
            try:
                if creator:
                    obj = creator(clazz)
                    if isinstance(obj, clazz):
                        return obj
                    raise TypeError("instance creator returned object of different type")
                return clazz()
            except Exception:
                log.exception("could not create pyro object instance")
                raise
        instance_mode, instance_creator = clazz._pyroInstancing
        if instance_mode == "single":
            # create and use one singleton instance of this class (not a global singleton, just exactly one per daemon)
            with self.create_single_instance_lock:
                instance = self._pyroInstances.get(clazz)
                if not instance:
                    log.debug("instancemode %s: creating new pyro object for %s", instance_mode, clazz)
                    instance = createInstance(clazz, instance_creator)
                    self._pyroInstances[clazz] = instance
                return instance
        elif instance_mode == "session":
            # Create and use one instance for this proxy connection
            # the instances are kept on the connection object.
            # (this is the default instance mode when using new style @expose)
            instance = conn.pyroInstances.get(clazz)
            if not instance:
                log.debug("instancemode %s: creating new pyro object for %s", instance_mode, clazz)
                instance = createInstance(clazz, instance_creator)
                conn.pyroInstances[clazz] = instance
            return instance
        elif instance_mode == "percall":
            # create and use a new instance just for this call
            log.debug("instancemode %s: creating new pyro object for %s", instance_mode, clazz)
            return createInstance(clazz, instance_creator)
        else:
            raise errors.DaemonError("invalid instancemode in registered class")

    def _sendExceptionResponse(self, connection, seq, serializer_id, exc_value, tbinfo, flags=0, annotations=None):
        """send an exception back including the local traceback info"""
        exc_value._pyroTraceback = tbinfo
        serializer = serializers.serializers_by_id[serializer_id]
        try:
            data = serializer.dumps(exc_value)
        except Exception:
            # the exception object couldn't be serialized, use a generic PyroError instead
            xt, xv, tb = sys.exc_info()
            msg = "Error serializing exception: %s. Original exception: %s: %s" % (str(xv), type(exc_value), str(exc_value))
            exc_value = errors.PyroError(msg)
            exc_value._pyroTraceback = tbinfo
            data = serializer.dumps(exc_value)
        flags |= protocol.FLAGS_EXCEPTION
        annotations = dict(annotations or {})
        annotations.update(self.annotations())
        msg = protocol.SendingMessage(protocol.MSG_RESULT, flags, seq, serializer.serializer_id, data, annotations=annotations)
        if config.LOGWIRE:
            protocol.log_wiredata(log, "daemon wiredata sending (error response)", msg)
        connection.send(msg.data)

    def register(self, obj_or_class, objectId=None, force=False, weak=False):
        """
        Register a Pyro object under the given id. Note that this object is now only
        known inside this daemon, it is not automatically available in a name server.
        This method returns a URI for the registered object.
        Pyro checks if an object is already registered, unless you set force=True.
        You can register a class or an object (instance) directly.
        For a class, Pyro will create instances of it to handle the remote calls according
        to the instance_mode (set via @expose on the class). The default there is one object
        per session (=proxy connection). If you register an object directly, Pyro will use
        that single object for *all* remote calls.
        With *weak=True*, only weak reference to the object will be stored, and the object will
        get unregistered from the daemon automatically when garbage-collected.
        """
        if objectId:
            if not isinstance(objectId, str):
                raise TypeError("objectId must be a string or None")
        else:
            objectId = "obj_" + uuid.uuid4().hex  # generate a new objectId
        if inspect.isclass(obj_or_class):
            if weak: raise TypeError("Classes cannot be registered with weak=True.")
            if not hasattr(obj_or_class, "_pyroInstancing"):
                obj_or_class._pyroInstancing = ("session", None)
        if not force:
            if hasattr(obj_or_class, "_pyroId") and obj_or_class._pyroId != "":  # check for empty string is needed for Cython
                pyro_id = obj_or_class._pyroId
                if pyro_id and self.objectsById.get(pyro_id) is obj_or_class:
                    raise errors.DaemonError("object or class already has a Pyro id")
            if objectId in self.objectsById:
                raise errors.DaemonError("an object or class is already registered with that id")
        # set some pyro attributes
        obj_or_class._pyroId = objectId
        obj_or_class._pyroDaemon = self
        # register a custom serializer for the type to automatically return proxies
        # we need to do this for all known serializers
        for ser in serializers.serializers.values():
            if inspect.isclass(obj_or_class):
                ser.register_type_replacement(obj_or_class, _pyro_obj_to_auto_proxy)
            else:
                ser.register_type_replacement(type(obj_or_class), _pyro_obj_to_auto_proxy)
        # register the object/class in the mapping
        self.objectsById[obj_or_class._pyroId] = obj_or_class if not weak else weakref.ref(obj_or_class)
        if weak: weakref.finalize(obj_or_class,self.unregister,objectId)
        return self.uriFor(objectId)

    def unregister(self, objectOrId):
        """
        Remove a class or object from the known objects inside this daemon.
        You can unregister the class/object directly, or with its id.
        """
        if objectOrId is None:
            raise ValueError("object or objectid argument expected")
        if not isinstance(objectOrId, str):
            objectId = getattr(objectOrId, "_pyroId", None)
            if objectId is None:
                raise errors.DaemonError("object isn't registered")
        else:
            objectId = objectOrId
            objectOrId = None
        if objectId == core.DAEMON_NAME:
            return
        if objectId in self.objectsById:
            del self.objectsById[objectId]
            if objectOrId is not None:
                del objectOrId._pyroId
                del objectOrId._pyroDaemon
                # Don't remove the custom type serializer because there may be
                # other registered objects of the same type still depending on it.

    def uriFor(self, objectOrId, nat=True):
        """
        Get a URI for the given object (or object id) from this daemon.
        Only a daemon can hand out proper uris because the access location is
        contained in them.
        Note that unregistered objects cannot be given an uri, but unregistered
        object names can (it's just a string we're creating in that case).
        If nat is set to False, the configured NAT address (if any) is ignored and it will
        return an URI for the internal address.
        """
        if not isinstance(objectOrId, str):
            objectOrId = getattr(objectOrId, "_pyroId", None)
            if objectOrId is None or objectOrId not in self.objectsById:
                raise errors.DaemonError("object isn't registered in this daemon")
        if nat:
            loc = self.natLocationStr or self.locationStr
        else:
            loc = self.locationStr
        return core.URI("PYRO:%s@%s" % (objectOrId, loc))

    def resetMetadataCache(self, objectOrId, nat=True):
        """Reset cache of metadata when a Daemon has available methods/attributes
        dynamically updated.  Clients will have to get a new proxy to see changes"""
        uri = self.uriFor(objectOrId, nat)
        # can only be cached if registered, else no-op
        if uri.object in self.objectsById:
            registered_object = _unpack_weakref(self.objectsById[uri.object])
            # Clear cache regardless of how it is accessed
            _reset_exposed_members(registered_object)

    def proxyFor(self, objectOrId, nat=True):
        """
        Get a fully initialized Pyro Proxy for the given object (or object id) for this daemon.
        If nat is False, the configured NAT address (if any) is ignored.
        The object or id must be registered in this daemon, or you'll get an exception.
        (you can't get a proxy for an unknown object)
        """
        uri = self.uriFor(objectOrId, nat)
        proxy = client.Proxy(uri)
        try:
            registered_object = _unpack_weakref(self.objectsById[uri.object])
        except KeyError:
            raise errors.DaemonError("object isn't registered in this daemon")
        meta = _get_exposed_members(registered_object)
        proxy._pyroGetMetadata(known_metadata=meta)
        return proxy

    def close(self):
        """Close down the server and release resources"""
        self.__mustshutdown.set()
        self.streaming_responses = {}
        if self.transportServer:
            log.debug("daemon closing")
            self.transportServer.close()
            self.transportServer = None

    def annotations(self):
        """Override to return a dict with custom user annotations to be sent with each response message."""
        return {}

    def combine(self, daemon):
        """
        Combines the event loop of the other daemon in the current daemon's loop.
        You can then simply run the current daemon's requestLoop to serve both daemons.
        This works fine on the multiplex server type, but doesn't work with the threaded server type.
        """
        log.debug("combining event loop with other daemon")
        self.transportServer.combine_loop(daemon.transportServer)

    def __annotations(self):
        annotations = current_context.response_annotations
        annotations.update(self.annotations())
        return annotations

    def __repr__(self):
        if hasattr(self, "locationStr"):
            family = socketutil.family_str(self.sock)
            return "<%s.%s at 0x%x; %s - %s; %d objects>" % (self.__class__.__module__, self.__class__.__name__,
                                                             id(self), self.locationStr, family, len(self.objectsById))
        else:
            # daemon objects may come back from serialized form without being properly initialized (by design)
            return "<%s.%s at 0x%x; unusable>" % (self.__class__.__module__, self.__class__.__name__, id(self))

    def __enter__(self):
        if not self.transportServer:
            raise errors.PyroError("cannot reuse this object")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __getstate__(self):
        # A little hack to make it possible to serialize Pyro objects, because they can reference a daemon,
        # but it is not meant to be able to properly serialize/deserialize Daemon objects.
        return tuple()

    def __setstate__(self, state):
        assert len(state) == 0

    def _streamResponse(self, data, client):
        if isinstance(data, collections.abc.Iterator) or inspect.isgenerator(data):
            if config.ITER_STREAMING:
                if type(data) in (type({}.keys()), type({}.values()), type({}.items())):
                    raise errors.PyroError("won't serialize or stream lazy dict iterators, convert to list yourself")
                stream_id = str(uuid.uuid4())
                self.streaming_responses[stream_id] = (client, time.time(), 0, data)
                return True, stream_id
            return True, None
        return False, data

    def __deserializeBlobArgs(self, protocolmsg):
        import marshal
        blobinfo = protocolmsg.annotations["BLBI"]
        blobinfo, objId, method = marshal.loads(blobinfo)
        blob = client.SerializedBlob(blobinfo, protocolmsg, is_blob=True)
        return objId, method, (blob,), {}  # object, method, vargs, kwargs


def serve(objects: Dict[Any, str], host: Optional[Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address]] = None,
          port: int = 0, daemon: Optional[Daemon] = None, use_ns: bool = True, verbose: bool = True) -> None:
    """
    Basic method to fire up a daemon (or supply one yourself).
    objects is a dict containing objects to register as keys, and
    their names (or None) as values. If ns is true they will be registered
    in the naming server as well, otherwise they just stay local.
    If you need to publish on a unix domain socket, or require finer control of the daemon's
    behavior, you can't use this shortcut method. Create a Daemon yourself and use its
    appropriate methods.
    See the documentation on 'publishing objects' (in chapter: Servers) for more details.
    """
    if daemon is None:
        daemon = Daemon(host, port)
    with daemon:
        ns = core.locate_ns() if use_ns else None
        for obj, name in objects.items():
            if ns:
                localname = None  # name is used for the name server
            else:
                localname = name  # no name server, use name in daemon
            uri = daemon.register(obj, localname)
            if verbose:
                print("Object {0}:\n    uri = {1}".format(repr(obj), uri))
            if name and ns:
                ns.register(name, uri)
                if verbose:
                    print("    name = {0}".format(name))
        if verbose:
            print("Pyro daemon running.")
        daemon.requestLoop()


def _default_methodcall_error_handler(daemon: Daemon, client_sock: socketutil.SocketConnection,
                                      method: Callable, vargs: Sequence[Any], kwargs: Dict[str, Any],
                                      exception: Exception) -> None:
    """The default routine called to process a exception raised in the user code of a method call"""
    log.debug("exception occurred in method call user code: client={} method={} exception={}"
              .format(client_sock, method.__qualname__, repr(exception)))


# register the special serializers for the pyro objects
serpent.register_class(Daemon, serializers.pyro_class_serpent_serializer)
serializers.SerializerBase.register_class_to_dict(Daemon, serializers.serialize_pyro_object_to_dict, serpent_too=False)


def _pyro_obj_to_auto_proxy(obj: Any) -> Any:
    """reduce function that automatically replaces Pyro objects by a Proxy"""
    daemon = getattr(obj, "_pyroDaemon", None)
    if daemon:
        # only return a proxy if the object is a registered pyro object
        return daemon.proxyFor(obj)
    return obj


def _get_attribute(obj: Any, attr: str) -> Any:
    """
    Resolves an attribute name to an object.  Raises
    an AttributeError if any attribute in the chain starts with a '``_``'.
    Doesn't resolve a dotted name, because that is a security vulnerability.
    It treats it as a single attribute name (and the lookup will likely fail).
    """
    if is_private_attribute(attr):
        raise AttributeError("attempt to access private attribute '%s'" % attr)
    else:
        obj = getattr(obj, attr)
    if getattr(obj, "_pyroExposed", False):
        return obj
    raise AttributeError("attempt to access unexposed attribute '%s'" % attr)


__exposed_member_cache = {}     # type: Dict[Tuple[type, bool], Dict[str, Set[str]]]


def _reset_exposed_members(obj: Any, only_exposed: bool = True) -> None:
    """Delete any cached exposed members forcing recalculation on next request"""
    if not inspect.isclass(obj):
        obj = obj.__class__
    cache_key = (obj, only_exposed)
    __exposed_member_cache.pop(cache_key, None)


def _get_exposed_members(obj: Any, only_exposed: bool = True) -> Dict[str, Set[str]]:
    """
    Return public and exposed members of the given object's class.
    You can also provide a class directly.
    Private members are ignored no matter what (names starting with underscore).
    If only_exposed is True, only members tagged with the @expose decorator are
    returned. If it is False, all public members are returned.
    The return value consists of the exposed methods, exposed attributes, and methods
    tagged as @oneway.
    (All this is used as meta data that Pyro sends to the proxy if it asks for it)
    """
    if not inspect.isclass(obj):
        obj = obj.__class__

    cache_key = (obj, only_exposed)
    if cache_key in __exposed_member_cache:
        return __exposed_member_cache[cache_key]

    methods = set()  # all methods
    oneway = set()  # oneway methods
    attrs = set()  # attributes
    for m in dir(obj):      # also lists names inherited from super classes
        if is_private_attribute(m):
            continue
        v = getattr(obj, m)
        if inspect.ismethod(v) or inspect.isfunction(v) or inspect.ismethoddescriptor(v):
            if getattr(v, "_pyroExposed", not only_exposed):
                methods.add(m)
                # check if the method is marked with the 'oneway' decorator:
                if getattr(v, "_pyroOneway", False):
                    oneway.add(m)
        elif inspect.isdatadescriptor(v):
            func = getattr(v, "fget", None) or getattr(v, "fset", None) or getattr(v, "fdel", None)
            if func is not None and getattr(func, "_pyroExposed", not only_exposed):
                attrs.add(m)
        # Note that we don't expose plain class attributes no matter what.
        # it is a syntax error to add a decorator on them, and it is not possible
        # to give them a _pyroExposed tag either.
        # The way to expose attributes is by using properties for them.
        # This automatically solves the protection/security issue: you have to
        # explicitly decide to make an attribute into a @property (and to @expose it)
        # before it becomes remotely accessible.
    result = {
        "methods": methods,
        "oneway": oneway,
        "attrs": attrs
    }
    __exposed_member_cache[cache_key] = result
    return result


def _unpack_weakref(obj: Any):
    """
    Unpack weak reference, or return the object itself, if not a weak reference.
    If the weak reference is dead (calling it returns None), raises an
    exception. Even though register(...,weak=True) creates finalizer which
    will delete the weakref from the mapping, it is possible that the object
    is garbage-collected asynchronously between obtaining weakref from the
    mapping and reference unpacking, making the weakref invalid; this is handled
    by the exception here.
    """
    if not isinstance(obj,weakref.ref): return obj
    ret=obj() # ret will hold strong reference to obj, until it gets deleted itself
    if ret is None: raise errors.DaemonError("Weakly registered deleted meanwhile (or finalizer failed?).")
    return ret

def _get_exposed_property_value(obj: Any, propname: str, only_exposed: bool = True) -> Any:
    """
    Return the value of an @exposed @property.
    If the requested property is not a @property or not exposed,
    an AttributeError is raised instead.
    """
    v = getattr(obj.__class__, propname)
    if inspect.isdatadescriptor(v):
        if v.fget and getattr(v.fget, "_pyroExposed", not only_exposed):
            return v.fget(obj)
    raise AttributeError("attempt to access unexposed or unknown remote attribute '%s'" % propname)


def _set_exposed_property_value(obj: Any, propname: str, value: Any, only_exposed: bool = True) -> Any:
    """
    Sets the value of an @exposed @property.
    If the requested property is not a @property or not exposed,
    an AttributeError is raised instead.
    """
    v = getattr(obj.__class__, propname)
    if inspect.isdatadescriptor(v):
        pfunc = v.fget or v.fset or v.fdel
        if v.fset and getattr(pfunc, "_pyroExposed", not only_exposed):
            return v.fset(obj, value)
    raise AttributeError("attempt to access unexposed or unknown remote attribute '%s'" % propname)


class _OnewayCallThread(threading.Thread):
    def __init__(self, pyro_method, vargs, kwargs, pyro_daemon, pyro_client_sock):
        super(_OnewayCallThread, self).__init__(target=self._methodcall, name="oneway-call")
        self.daemon = True
        self.parent_context = current_context.to_global()
        self.pyro_daemon = pyro_daemon
        self.pyro_client_sock = pyro_client_sock
        self.pyro_method = pyro_method
        self.pyro_vargs = vargs
        self.pyro_kwars = kwargs

    def run(self):
        current_context.from_global(self.parent_context)
        super(_OnewayCallThread, self).run()

    def _methodcall(self):
        try:
            self.pyro_method(*self.pyro_vargs, **self.pyro_kwars)
        except Exception as xv:
            self.pyro_daemon.methodcall_error_handler(self.pyro_daemon, self.pyro_client_sock,
                                                      self.pyro_method, self.pyro_vargs, self.pyro_kwars,
                                                      xv)



"""
Tests for the http gateway.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import pytest
import json
from wsgiref.util import setup_testing_defaults
import io
import Pyro5.utils.httpgateway
import Pyro5.errors
import Pyro5.core
from Pyro5.nameserver import NameServer


class WSGITestBase:
    """Helper class for wsgi unit-tests. Provides up a simple interface to make requests
    as though they came through a wsgi interface from a user."""

    def __init__(self):
        """Set up a fresh testing environment before each test."""
        self.cookies = []

    def request(self, application, url, query_string="", post_data=b""):
        """Hand a request to the application as if sent by a client.
        @param application: The callable wsgi application to test.
        @param url: The URL to make the request against.
        @param query_string: Url parameters.
        @param post_data: bytes to post."""
        self.response_started = False
        method = 'POST' if post_data else 'GET'
        temp = io.BytesIO(post_data)
        environ = {
            'PATH_INFO': url,
            'REQUEST_METHOD': method,
            'CONTENT_LENGTH': len(post_data),
            'QUERY_STRING': query_string,
            'wsgi.input': temp,
        }
        if method == "POST":
            environ["CONTENT_TYPE"] = "application/x-www-form-urlencoded"
        setup_testing_defaults(environ)
        if self.cookies:
            environ['HTTP_COOKIE'] = ';'.join(self.cookies)
        response = b''
        for ret in application(environ, self._start_response):
            assert self.response_started
            response += ret
        temp.close()
        return response

    def _start_response(self, status, headers):
        """A callback passed into the application, to simulate a wsgi
        environment.

        @param status: The response status of the application ("200", "404", etc)
        @param headers: Any headers to begin the response with.
        """
        assert not self.response_started
        self.response_started = True
        self.status = status
        self.headers = headers
        for header in headers:
            # Parse out any cookies and save them to send with later requests.
            if header[0] == 'Set-Cookie':
                var = header[1].split(';', 1)
                if len(var) > 1 and var[1][0:9] == ' Max-Age=':
                    if int(var[1][9:]) > 0:
                        # An approximation, since our cookies never expire unless
                        # explicitly deleted (by setting Max-Age=0).
                        self.cookies.append(var[0])
                    else:
                        index = self.cookies.index(var[0])
                        self.cookies.pop(index)

    def new_session(self):
        """Start a new session (or pretend to be a different user) by deleting
        all current cookies."""
        self.cookies = []


@pytest.fixture(scope="module")
def wsgiserver():
    # a bit of hackery to avoid having to launch a live name server
    class NameServerDummyProxy(NameServer):
        def __init__(self):
            super(NameServerDummyProxy, self).__init__()
            self._pyroUri = Pyro5.core.URI("PYRO:dummy12345@localhost:59999")
            self.register("http.ObjectName", "PYRO:dummy12345@localhost:59999")

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def __call__(self, *args, **kwargs):
            return ["Name1", "Name2", "Name3"]

        def _pyroInvokeBatch(self, calls, oneway=False):
            return ["Name1"]

        def _pyroClaimOwnership(self):
            pass

    ws = WSGITestBase()
    old_get_ns = Pyro5.utils.httpgateway.get_nameserver
    Pyro5.utils.httpgateway.get_nameserver = lambda: NameServerDummyProxy()
    Pyro5.config.COMMTIMEOUT = 0.3
    yield ws
    Pyro5.utils.httpgateway.get_nameserver = old_get_ns
    Pyro5.config.COMMTIMEOUT = 0.0


class TestHttpGateway:

    def teardown_class(self):
        Pyro5.config.SERIALIZER = "serpent"

    def test_params(self):
        multiparams = {
            "first": [1],
            "second": [1, 2, 3],
            "third": 42
        }
        checkparams = {
            "first": 1,
            "second": [1, 2, 3],
            "third": 42
        }
        params = Pyro5.utils.httpgateway.singlyfy_parameters(multiparams)
        assert checkparams == params
        params = Pyro5.utils.httpgateway.singlyfy_parameters(multiparams)
        assert checkparams == params

    def test_redirect(self, wsgiserver):
        result = wsgiserver.request(Pyro5.utils.httpgateway.pyro_app, "/")
        assert wsgiserver.status == "302 Found"
        assert wsgiserver.headers == [('Location', '/pyro/')]
        assert result == b""

    def test_webpage(self, wsgiserver):
        result = wsgiserver.request(Pyro5.utils.httpgateway.pyro_app, "/pyro/")
        assert wsgiserver.status == "200 OK"
        assert result.startswith(b"<!DOCTYPE html>")
        assert len(result) > 1000

    def test_methodCallGET(self, wsgiserver):
        result = wsgiserver.request(Pyro5.utils.httpgateway.pyro_app, "/pyro/http.ObjectName/method", query_string="param=42&param2=hello")
        # the call will result in a communication error because the dummy uri points to something that is not available
        assert wsgiserver.status == "500 Internal Server Error"
        j = json.loads(result.decode("utf-8"))
        assert j["__exception__"]
        assert j["__class__"] == "Pyro5.errors.CommunicationError"

    def test_methodCallPOST(self, wsgiserver):
        result = wsgiserver.request(Pyro5.utils.httpgateway.pyro_app, "/pyro/http.ObjectName/method", post_data=b"param=42&param2=hello")
        # the call will result in a communication error because the dummy uri points to something that is not available
        assert wsgiserver.status == "500 Internal Server Error"
        j = json.loads(result.decode("utf-8"))
        assert j["__exception__"]
        assert j["__class__"] == "Pyro5.errors.CommunicationError"

    def test_nameDeniedPattern(self, wsgiserver):
        result = wsgiserver.request(Pyro5.utils.httpgateway.pyro_app, "/pyro/Pyro.NameServer/method")
        # the call will result in a access denied error because the uri points to a non-exposed name
        assert wsgiserver.status == "403 Forbidden"

    def test_nameDeniedNotRegistered(self, wsgiserver):
        result = wsgiserver.request(Pyro5.utils.httpgateway.pyro_app, "/pyro/http.NotRegisteredName/method")
        # the call will result in a communication error because the dummy uri points to something that is not registered
        assert wsgiserver.status == "500 Internal Server Error"
        j = json.loads(result.decode("utf-8"))
        assert j["__exception__"]
        assert j["__class__"] == "Pyro5.errors.NamingError"

    def test_exposedPattern(self):
        assert Pyro5.utils.httpgateway.pyro_app.ns_regex == r"http\."



import logging
from Pyro5.api import expose, serve, config


logging.basicConfig(level=logging.DEBUG)
logging.getLogger("Pyro5").setLevel(logging.DEBUG)

config.COMMTIMEOUT = 5.0
config.POLLTIMEOUT = 5.0  # only used for multiplexing server


class TestDisconnect(object):
    @expose
    def echo(self, arg):
        print("echo: ", arg)
        return arg


serve({
    TestDisconnect: "example.disconnect"
}, use_ns=False)



import Pyro5.api
import chain


this_node = "A"
next_node = "B"

servername = "example.chain." + this_node

with Pyro5.api.Daemon() as daemon:
    obj = chain.Chain(this_node, next_node)
    uri = daemon.register(obj)
    with Pyro5.api.locate_ns() as ns:
        ns.register(servername, uri)

    # enter the service loop.
    print("Server started %s" % this_node)
    daemon.requestLoop()



"""
Tests for the built-in test echo server.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import time
import pytest
from threading import Thread, Event
import Pyro5.client
import Pyro5.errors
import Pyro5.utils.echoserver as echoserver
from Pyro5 import config


class EchoServerThread(Thread):
    def __init__(self):
        super(EchoServerThread, self).__init__()
        self.daemon = True
        self.started = Event()
        self.echodaemon = self.echoserver = self.uri = None

    def run(self):
        self.echodaemon, self.echoserver, self.uri = echoserver.main(args=["-q"], returnWithoutLooping=True)
        self.started.set()
        self.echodaemon.requestLoop(loopCondition=lambda: not self.echoserver._must_shutdown)


class TestEchoserver:
    def setup_method(self):
        self.echoserverthread = EchoServerThread()
        self.echoserverthread.start()
        self.echoserverthread.started.wait()
        self.uri = self.echoserverthread.uri

    def teardown_method(self):
        self.echoserverthread.echodaemon.shutdown()
        time.sleep(0.02)
        self.echoserverthread.join()
        config.SERVERTYPE = "thread"

    def testExposed(self):
        e = Pyro5.utils.echoserver.EchoServer()
        assert hasattr(e, "_pyroExposed")

    def testEcho(self):
        with Pyro5.client.Proxy(self.uri) as echo:
            try:
                assert echo.echo("hello") == "hello"
                assert echo.echo(None) is None
                assert echo.echo([1,2,3]) == [1,2,3]
            finally:
                echo.shutdown()

    def testError(self):
        with Pyro5.client.Proxy(self.uri) as echo:
            with pytest.raises(Exception) as x:
                echo.error()
            tb = "".join(Pyro5.errors.get_pyro_traceback(x.type, x.value, x.tb))
            assert "Remote traceback" in tb
            assert "ValueError" in tb
            assert str(x.value) == "this is the generated error from echoserver echo() method"
            with pytest.raises(Exception) as x:
                echo.error_with_text()
            tb = "".join(Pyro5.errors.get_pyro_traceback(x.type, x.value, x.tb))
            assert "Remote traceback" in tb
            assert "ValueError" in tb
            assert str(x.value) == "the message of the error"

    def testGenerator(self):
        with Pyro5.client.Proxy(self.uri) as echo:
            remotegenerator = echo.generator()
            assert isinstance(remotegenerator, Pyro5.client._StreamResultIterator)
            next(remotegenerator)
            next(remotegenerator)
            next(remotegenerator)
            with pytest.raises(StopIteration):
                next(remotegenerator)




import Pyro5.api


uri = input("Enter the uri of the dmesg server: ")
dmesg = Pyro5.api.Proxy(uri)
try:
    print("Last few lines of the dmesg kernel buffer:\n")
    lines = dmesg.dmesg()
    print("\n".join(lines))
    print("\nNormally you can't read this info from a non-root user. "
          "But the server is running as root and is able to access it for you.")
    print("TIP: now kill the server if you no longer need it!")
except Exception as x:
    print("ERROR:", x)
    print("Is the server running with root privileges?")



"""
Socket server based on a worker thread pool. Doesn't use select.

Uses a single worker thread per client connection.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import socket
import logging
import sys
import time
import threading
import os
import selectors
import contextlib
from . import config, socketutil, errors

log = logging.getLogger("Pyro5.threadpoolserver")
_client_disconnect_lock = threading.Lock()


class ClientConnectionJob(object):
    """
    Takes care of a single client connection and all requests
    that may arrive during its life span.
    """

    def __init__(self, clientSocket, clientAddr, daemon):
        self.csock = socketutil.SocketConnection(clientSocket)
        self.caddr = clientAddr
        self.daemon = daemon

    def __call__(self):
        if self.handleConnection():
            try:
                while True:
                    try:
                        self.daemon.handleRequest(self.csock)
                    except (socket.error, errors.ConnectionClosedError):
                        # client went away.
                        log.debug("disconnected %s", self.caddr)
                        break
                    except errors.SecurityError:
                        log.debug("security error on client %s", self.caddr)
                        break
                    except errors.TimeoutError as x:
                        # for timeout errors we're not really interested in detailed traceback info
                        log.warning("error during handleRequest: %s" % x)
                        break
                    except Exception:
                        # other errors log a warning, break this loop and close the client connection
                        ex_t, ex_v, ex_tb = sys.exc_info()
                        tb = errors.format_traceback(ex_t, ex_v, ex_tb)
                        msg = "error during handleRequest: %s; %s" % (ex_v, "".join(tb))
                        log.warning(msg)
                        break
            finally:
                with _client_disconnect_lock:
                    try:
                        self.daemon._clientDisconnect(self.csock)
                    except Exception as x:
                        log.warning("Error in clientDisconnect: " + str(x))
                self.csock.close()

    def handleConnection(self):
        # connection handshake
        try:
            if self.daemon._handshake(self.csock):
                return True
            self.csock.close()
        except Exception:
            ex_t, ex_v, ex_tb = sys.exc_info()
            tb = errors.format_traceback(ex_t, ex_v, ex_tb)
            log.warning("error during connect/handshake: %s; %s", ex_v, "\n".join(tb))
            self.csock.close()
        return False

    def denyConnection(self, reason):
        log.warning("client connection was denied: " + reason)
        # return failed handshake
        self.daemon._handshake(self.csock, denied_reason=reason)
        self.csock.close()


class Housekeeper(threading.Thread):
    def __init__(self, daemon):
        super(Housekeeper, self).__init__(name="housekeeper")
        self.pyroDaemon = daemon
        self.stop = threading.Event()
        self.daemon = True
        self.waittime = min(config.POLLTIMEOUT or 0, max(config.COMMTIMEOUT or 0, 5))

    def run(self):
        while True:
            if self.stop.wait(self.waittime):
                break
            self.pyroDaemon._housekeeping()


class SocketServer_Threadpool(object):
    """transport server for socket connections, worker thread pool version."""

    def __init__(self):
        self.daemon = self.sock = self._socketaddr = self.locationStr = self.pool = None
        self.shutting_down = False
        self.housekeeper = None
        self._selector = selectors.DefaultSelector()

    def init(self, daemon, host, port, unixsocket=None):
        log.info("starting thread pool socketserver")
        self.daemon = daemon
        self.sock = None
        bind_location = unixsocket if unixsocket else (host, port)
        if config.SSL:
            sslContext = socketutil.get_ssl_context(servercert=config.SSL_SERVERCERT,
                                                    serverkey=config.SSL_SERVERKEY,
                                                    keypassword=config.SSL_SERVERKEYPASSWD,
                                                    cacerts=config.SSL_CACERTS)
            log.info("using SSL,  cert=%s  key=%s  cacerts=%s", config.SSL_SERVERCERT, config.SSL_SERVERKEY, config.SSL_CACERTS)
        else:
            sslContext = None
            log.info("not using SSL")
        self.sock = socketutil.create_socket(bind=bind_location,
                                             reuseaddr=config.SOCK_REUSE,
                                             timeout=config.COMMTIMEOUT,
                                             noinherit=True,
                                             nodelay=config.SOCK_NODELAY,
                                             sslContext=sslContext)
        self._socketaddr = self.sock.getsockname()
        if not unixsocket and self._socketaddr[0].startswith("127."):
            if host is None or host.lower() != "localhost" and not host.startswith("127."):
                log.warning("weird DNS setup: %s resolves to localhost (127.x.x.x)", host)
        if unixsocket:
            self.locationStr = "./u:" + unixsocket
        else:
            host = host or self._socketaddr[0]
            port = port or self._socketaddr[1]
            if ":" in host:  # ipv6
                self.locationStr = "[%s]:%d" % (host, port)
            else:
                self.locationStr = "%s:%d" % (host, port)
        self.pool = Pool()
        self.housekeeper = Housekeeper(daemon)
        self.housekeeper.start()
        self._selector.register(self.sock, selectors.EVENT_READ, self)

    def __del__(self):
        if self.sock is not None:
            self.sock.close()
            self.sock = None
        if self.pool is not None:
            self.pool.close()
            self.pool = None
        if self.housekeeper:
            self.housekeeper.stop.set()
            self.housekeeper.join()
            self.housekeeper = None

    def __repr__(self):
        return "<%s on %s; %d workers>" % (self.__class__.__name__, self.locationStr, self.pool.num_workers())

    def loop(self, loopCondition=lambda: True):
        log.debug("threadpool server requestloop")
        while (self.sock is not None) and not self.shutting_down and loopCondition():
            try:
                self.events([self.sock])
            except (socket.error, OSError) as x:
                if not loopCondition():
                    # swallow the socket error if loop terminates anyway
                    # this can occur if we are asked to shutdown, socket can be invalid then
                    break
                # socket errors may not lead to a server abort, so we log it and continue
                err = getattr(x, "errno", x.args[0])
                log.warning("socket error '%s' with errno=%d, shouldn't happen", x, err)
                continue
            except KeyboardInterrupt:
                log.debug("stopping on break signal")
                break

    def combine_loop(self, server):
        raise TypeError("You can't use the loop combiner on the threadpool server type")

    def events(self, eventsockets):
        """used for external event loops: handle events that occur on one of the sockets of this server"""
        # we only react on events on our own server socket.
        # all other (client) sockets are owned by their individual threads.
        assert self.sock in eventsockets
        with contextlib.suppress(socket.timeout):   # just continue the loop on a timeout on accept
            events = self._selector.select(config.POLLTIMEOUT)
            if not events:
                return
            csock, caddr = self.sock.accept()
            if self.shutting_down:
                csock.close()
                return
            if hasattr(csock, "getpeercert"):
                log.debug("connected %s - SSL", caddr)
            else:
                log.debug("connected %s - unencrypted", caddr)
            if config.COMMTIMEOUT:
                csock.settimeout(config.COMMTIMEOUT)
            job = ClientConnectionJob(csock, caddr, self.daemon)
            try:
                self.pool.process(job)
            except NoFreeWorkersError:
                job.denyConnection("no free workers, increase server threadpool size")

    def shutdown(self):
        self.shutting_down = True
        self.wakeup()
        time.sleep(0.05)
        self.close()
        self.sock = None

    def close(self):
        if self.housekeeper:
            self.housekeeper.stop.set()
            self.housekeeper.join()
            self.housekeeper = None
        if self.sock:
            with contextlib.suppress(socket.error, OSError):
                sockname = self.sock.getsockname()
            with contextlib.suppress(Exception):
                self.sock.close()
                if type(sockname) is str:
                    # it was a Unix domain socket, remove it from the filesystem
                    if os.path.exists(sockname):
                        os.remove(sockname)
            self.sock = None
        self.pool.close()

    @property
    def sockets(self):
        # the server socket is all we care about, all client sockets are running in their own threads
        return [self.sock]

    @property
    def selector(self):
        raise TypeError("threadpool server doesn't have multiplexing selector")

    def wakeup(self):
        socketutil.interrupt_socket(self._socketaddr)


class PoolError(Exception):
    pass


class NoFreeWorkersError(PoolError):
    pass


class Worker(threading.Thread):
    def __init__(self, pool):
        super(Worker, self).__init__()
        self.daemon = True
        self.name = "Pyro-Worker-%d" % id(self)
        self.job_available = threading.Event()
        self.job = None
        self.pool = pool

    def process(self, job):
        self.job = job
        self.job_available.set()

    def run(self):
        while True:
            self.job_available.wait()
            self.job_available.clear()
            if self.job is None:
                break
            try:
                self.job()
            except Exception as x:
                log.exception("unhandled exception from job in worker thread %s: %s", self.name, x)
            self.job = None
            self.pool.notify_done(self)
        self.pool = None


class Pool(object):
    """
    A job processing pool that is using a pool of worker threads.
    The amount of worker threads in the pool is configurable and scales between min/max size.
    """
    def __init__(self):
        if config.THREADPOOL_SIZE < 1 or config.THREADPOOL_SIZE_MIN < 1:
            raise ValueError("threadpool sizes must be greater than zero")
        if config.THREADPOOL_SIZE_MIN > config.THREADPOOL_SIZE:
            raise ValueError("minimum threadpool size must be less than or equal to max size")
        self.idle = set()
        self.busy = set()
        self.closed = False
        for _ in range(config.THREADPOOL_SIZE_MIN):
            worker = Worker(self)
            self.idle.add(worker)
            worker.start()
        log.debug("worker pool created with initial size %d", self.num_workers())
        self.count_lock = threading.Lock()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if not self.closed:
            log.debug("closing down")
            for w in list(self.busy):
                w.process(None)
            for w in list(self.idle):
                w.process(None)
            self.closed = True
            time.sleep(0.1)
            idle, self.idle = self.idle, set()
            busy, self.busy = self.busy, set()
            # check if the threads that are joined are not the current thread.
            current_thread = threading.current_thread()
            while idle:
                p = idle.pop()
                if p is not current_thread:
                    p.join(timeout=0.1)
            while busy:
                p = busy.pop()
                if p is not current_thread:
                    p.join(timeout=0.1)

    def __repr__(self):
        return "<%s.%s at 0x%x; %d busy workers; %d idle workers>" % \
               (self.__class__.__module__, self.__class__.__name__, id(self), len(self.busy), len(self.idle))

    def num_workers(self):
        return len(self.busy) + len(self.idle)

    def process(self, job):
        if self.closed:
            raise PoolError("job queue is closed")
        if self.idle:
            worker = self.idle.pop()
        elif self.num_workers() < config.THREADPOOL_SIZE:
            worker = Worker(self)
            worker.start()
        else:
            raise NoFreeWorkersError("no free workers available, increase thread pool size")
        self.busy.add(worker)
        worker.process(job)
        log.debug("worker counts: %d busy, %d idle", len(self.busy), len(self.idle))

    def notify_done(self, worker):
        if worker in self.busy:
            self.busy.remove(worker)
        if self.closed:
            worker.process(None)
            return
        if len(self.idle) >= config.THREADPOOL_SIZE_MIN:
            worker.process(None)
        else:
            self.idle.add(worker)
        log.debug("worker counts: %d busy, %d idle", len(self.busy), len(self.idle))



"""
An effort to provide a backward-compatible Pyro4 API layer,
to make porting existing code from Pyro4 to Pyro5 easier.

This only works for code that imported Pyro4 symbols from the Pyro4 module
directly, instead of from one of Pyro4's sub modules. So, for instance:
from Pyro4 import Proxy  instead of:  from Pyro4.core import Proxy

*some* submodules are more or less emulated such as Pyro4.errors, Pyro4.socketutil.

So, you may first have to convert your old code to use the importing scheme to
only import the Pyro4 module and not from its submodules, and then you should
insert this at the top to enable the compatibility layer:
from Pyro5.compatibility import Pyro4


Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import sys
import ipaddress

from .. import api
from .. import errors
from .. import serializers
from .. import socketutil as socketutil_pyro5
from ..configure import Configuration


__all__ = ["config", "URI", "Proxy", "Daemon", "callback", "batch",
           "asyncproxy", "oneway", "expose", "behavior", "current_context",
           "locateNS", "resolve", "Future", "errors"]


# symbols that are no longer available in Pyro5 and that we don't emulate:

def asyncproxy(*args, **kwargs):
    raise NotImplementedError("async proxy is no longer available in Pyro5")


class Future(object):
    def __init__(self, *args):
        raise NotImplementedError("Pyro5 no longer provides its own Future class, "
                                  "you should use Python's concurrent.futures module instead for that")


class NamespaceInterceptor:
    def __init__(self, namespace):
        self.namespace = namespace

    def __getattr__(self, item):
        raise NotImplementedError("The Pyro4 compatibility layer doesn't provide the Pyro4.{0} module, "
                                  "first make sure the code only uses symbols from the Pyro4 package directly"
                                  .format(self.namespace))


naming = NamespaceInterceptor("naming")
core = NamespaceInterceptor("core")
message = NamespaceInterceptor("message")


# compatibility wrappers for the other symbols:

__version__ = api.__version__
callback = api.callback
oneway = api.oneway
expose = api.expose
behavior = api.behavior
current_context = api.current_context


class CompatConfiguration(Configuration):
    def asDict(self):
        return self.as_dict()


config = CompatConfiguration()


class URI(api.URI):
    pass


class Proxy(api.Proxy):
    def _pyroAsync(self, asynchronous=True):
        raise NotImplementedError("async proxy is no longer available in Pyro5")

    @property
    def _pyroHmacKey(self):
        raise NotImplementedError("pyro5 doesn't have hmacs anymore")

    def __setattr__(self, name, value):
        if name == "_pyroHmacKey":
            raise NotImplementedError("pyro5 doesn't have hmacs anymore")
        return super().__setattr__(name, value)


class Daemon(api.Daemon):
    pass


def locateNS(host=None, port=None, broadcast=True, hmac_key=None):
    if hmac_key:
        raise NotImplementedError("hmac_key is no longer available in Pyro5, consider using 2-way SSL instead")
    return api.locate_ns(host, port, broadcast)


def resolve(uri, hmac_key=None):
    if hmac_key:
        raise NotImplementedError("hmac_key is no longer available in Pyro5, consider using 2-way SSL instead")
    return api.resolve(uri)


class BatchProxy(api.BatchProxy):
    def __call__(self, oneway=False, asynchronous=False):
        if asynchronous:
            raise NotImplementedError("async proxy is no longer available in Pyro5")
        return super().__call__(oneway)


def batch(proxy):
    return BatchProxy(proxy)


class UtilModule:
    @staticmethod
    def getPyroTraceback(ex_type=None, ex_value=None, ex_tb=None):
        return errors.get_pyro_traceback(ex_type, ex_value, ex_tb)

    @staticmethod
    def formatTraceback(ex_type=None, ex_value=None, ex_tb=None, detailed=False):
        return errors.format_traceback(ex_type, ex_value, ex_tb, detailed)

    SerializerBase = serializers.SerializerBase

    def excepthook(self, *args, **kwargs):
        return errors.excepthook(*args, **kwargs)


util = UtilModule()


class SocketUtilModule:
    @staticmethod
    def getIpVersion(hostnameOrAddress):
        return ipaddress.ip_address(hostnameOrAddress).version

    @staticmethod
    def getIpAddress(hostname, workaround127=False, ipVersion=None):
        return str(socketutil_pyro5.get_ip_address(hostname, workaround127, ipVersion))

    @staticmethod
    def getInterfaceAddress(ip_address):
        return str(socketutil_pyro5.get_interface(ip_address).ip)

    @staticmethod
    def createSocket(bind=None, connect=None, reuseaddr=False, keepalive=True,
                     timeout=-1, noinherit=False,
                     ipv6=False, nodelay=True, sslContext=None):
        return socketutil_pyro5.create_socket(bind, connect, reuseaddr, keepalive,
                                              timeout, noinherit, ipv6, nodelay, sslContext)

    @staticmethod
    def createBroadcastSocket(bind=None, reuseaddr=False, timeout=-1, ipv6=False):
        return socketutil_pyro5.create_bc_socket(bind, reuseaddr, timeout, ipv6)

    @staticmethod
    def receiveData(sock, size):
        return socketutil_pyro5.receive_data(sock, size)

    @staticmethod
    def sendData(sock, data):
        return socketutil_pyro5.send_data(sock, data)


socketutil = SocketUtilModule()


class ConstantsModule:
    from .. import __version__ as VERSION
    from ..core import DAEMON_NAME, NAMESERVER_NAME
    from ..protocol import PROTOCOL_VERSION


constants = ConstantsModule()


# make sure that subsequent  from Pyro4 import ...  will work:
sys.modules["Pyro4"] = sys.modules[__name__]
sys.modules["Pyro4.errors"] = errors
sys.modules["Pyro4.core"] = core
sys.modules["Pyro4.naming"] = naming
sys.modules["Pyro4.util"] = util
sys.modules["Pyro4.socketutil"] = socketutil
sys.modules["Pyro4.constants"] = constants
sys.modules["Pyro4.message"] = message



import os
import pwd
import Pyro5.api


class RestrictedService:
    @Pyro5.api.expose
    def who_is_server(self):
        return os.getuid(), os.getgid(), pwd.getpwuid(os.getuid()).pw_name

    @Pyro5.api.expose
    def write_file(self):
        # this should fail ("permission denied") because of the dropped privileges
        with open("dummy-test-file.bin", "w"):
            pass


class RestrictedDaemon(Pyro5.api.Daemon):
    def __init__(self):
        super().__init__()
        print("Server started as:")
        print(" uid/gid", os.getuid(), os.getgid())
        print(" euid/egid", os.geteuid(), os.getegid())
        self.drop_privileges("nobody")

    def drop_privileges(self, user):
        nobody = pwd.getpwnam(user)
        try:
            os.setgid(nobody.pw_uid)
            os.setuid(nobody.pw_gid)
        except OSError:
            print("Failed to drop privileges. You'll have to start this program as root to be able to do this.")
            raise
        print("Privileges dropped. Server now running as", user)
        print(" uid/gid", os.getuid(), os.getgid())
        print(" euid/egid", os.geteuid(), os.getegid())


if __name__ == "__main__":
    rdaemon = RestrictedDaemon()
    Pyro5.api.serve({
        RestrictedService: "restricted"
    }, host="localhost", daemon=rdaemon, use_ns=False)



"""
Tests for the daemon.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import time
import socket
import uuid
import pytest
import Pyro5.core
import Pyro5.client
import Pyro5.server
import Pyro5.nameserver
import Pyro5.protocol
import Pyro5.socketutil
import Pyro5.serializers
from Pyro5.errors import DaemonError, PyroError
from Pyro5 import config
from Pyro5.callcontext import current_context
from support import *


class MyObj(object):
    def __init__(self, arg):
        self.arg = arg

    def __eq__(self, other):
        return self.arg == other.arg

    __hash__ = object.__hash__


class CustomDaemonInterface(Pyro5.server.DaemonObject):
    def __init__(self, daemon):
        super(CustomDaemonInterface, self).__init__(daemon)

    def custom_daemon_method(self):
        return 42


class TestDaemon:
    # We create a daemon, but notice that we are not actually running the requestloop.
    # 'on-line' tests are all taking place in another test, to keep this one simple.

    def setUp(self):
        config.POLLTIMEOUT = 0.1

    def sendHandshakeMessage(self, conn, correlation_id=None):
        ser = Pyro5.serializers.serializers_by_id[Pyro5.serializers.MarshalSerializer.serializer_id]
        data = ser.dumps({"handshake": "hello", "object": Pyro5.core.DAEMON_NAME})
        current_context.correlation_id = correlation_id
        msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_CONNECT, 0, 99, Pyro5.serializers.MarshalSerializer.serializer_id, data)
        conn.send(msg.data)

    def testSerializerAccepted(self):
        with Pyro5.server.Daemon(port=0) as d:
            msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 0, Pyro5.serializers.MarshalSerializer.serializer_id, b"")
            cm = ConnectionMock(msg)
            d.handleRequest(cm)
            msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 0, Pyro5.serializers.JsonSerializer.serializer_id, b"")
            cm = ConnectionMock(msg)
            d.handleRequest(cm)
            msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, b"")
            cm = ConnectionMock(msg)
            d.handleRequest(cm)
            if "msgpack" in Pyro5.serializers.serializers:
                msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 0, Pyro5.serializers.MsgpackSerializer.serializer_id, b"")
                cm = ConnectionMock(msg)
                d.handleRequest(cm)

    def testDaemon(self):
        with Pyro5.server.Daemon(port=0) as d:
            hostname, port = d.locationStr.split(":")
            port = int(port)
            assert Pyro5.core.DAEMON_NAME in d.objectsById
            assert str(d.uriFor(Pyro5.core.DAEMON_NAME)) == "PYRO:" + Pyro5.core.DAEMON_NAME + "@" + d.locationStr
            # check the string representations
            expected = "<Pyro5.server.Daemon at 0x%x; %s - %s; 1 objects>" % (id(d), d.locationStr, Pyro5.socketutil.family_str(d.sock))
            assert str(d) == expected
            assert repr(d) == expected
            sockname = d.sock.getsockname()
            assert sockname[1] == port
            daemonobj = d.objectsById[Pyro5.core.DAEMON_NAME]
            daemonobj.ping()
            daemonobj.registered()

    def testDaemonCustomInterface(self):
        with Pyro5.server.Daemon(port=0, interface=CustomDaemonInterface) as d:
            obj = d.objectsById[Pyro5.core.DAEMON_NAME]
            assert obj.custom_daemon_method() == 42

    def testDaemonConnectedSocket(self):
        try:
            Pyro5.config.SERVERTYPE = "thread"
            with Pyro5.server.Daemon() as d:
                assert "Thread" in d.transportServer.__class__.__name__
            s1, s2 = socket.socketpair()
            with Pyro5.server.Daemon(connected_socket=s1) as d:
                assert d.locationStr=="./u:<<not-bound>>" or d.locationStr.startswith("127.0.")
                assert not("Thread" in d.transportServer.__class__.__name__)
                assert "Existing" in d.transportServer.__class__.__name__
            Pyro5.config.SERVERTYPE = "multiplex"
            with Pyro5.server.Daemon() as d:
                assert "Multiplex" in d.transportServer.__class__.__name__
            s1, s2 = socket.socketpair()
            with Pyro5.server.Daemon(connected_socket=s1) as d:
                assert d.locationStr=="./u:<<not-bound>>" or d.locationStr.startswith("127.0.")
                assert not("Multiplex" in d.transportServer.__class__.__name__)
                assert "Existing" in d.transportServer.__class__.__name__
        finally:
            Pyro5.config.SERVERTYPE = "thread"

    def testDaemonUnixSocket(self):
        if hasattr(socket, "AF_UNIX"):
            SOCKNAME = "test_unixsocket"
            with Pyro5.server.Daemon(unixsocket=SOCKNAME) as d:
                locationstr = "./u:" + SOCKNAME
                assert d.locationStr == locationstr
                assert str(d.uriFor(Pyro5.core.DAEMON_NAME)) == "PYRO:" + Pyro5.core.DAEMON_NAME + "@" + locationstr
                # check the string representations
                expected = "<Pyro5.server.Daemon at 0x%x; %s - Unix; 1 objects>" % (id(d), locationstr)
                assert str(d) == expected
                assert d.sock.getsockname() == SOCKNAME
                assert d.sock.family == socket.AF_UNIX

    def testDaemonUnixSocketAbstractNS(self):
        if hasattr(socket, "AF_UNIX"):
            SOCKNAME = "\0test_unixsocket"  # mind the \0 at the start, for a Linux abstract namespace socket
            with Pyro5.server.Daemon(unixsocket=SOCKNAME) as d:
                locationstr = "./u:" + SOCKNAME
                assert d.locationStr == locationstr
                assert str(d.uriFor(Pyro5.core.DAEMON_NAME)) == "PYRO:" + Pyro5.core.DAEMON_NAME + "@" + locationstr
                # check the string representations
                expected = "<Pyro5.server.Daemon at 0x%x; %s - Unix; 1 objects>" % (id(d), locationstr)
                assert str(d) == expected
                sn_bytes = bytes(SOCKNAME, "ascii")
                assert d.sock.getsockname() == sn_bytes
                assert d.sock.family == socket.AF_UNIX

    def testServertypeThread(self):
        old_servertype = config.SERVERTYPE
        config.SERVERTYPE = "thread"
        with Pyro5.server.Daemon(port=0) as d:
            assert d.sock in d.sockets, "daemon's socketlist should contain the server socket"
            assert len(d.sockets) == 1, "daemon without connections should have just 1 socket"
        config.SERVERTYPE = old_servertype

    def testServertypeMultiplex(self):
        old_servertype = config.SERVERTYPE
        config.SERVERTYPE = "multiplex"
        with Pyro5.server.Daemon(port=0) as d:
            assert d.sock in d.sockets, "daemon's socketlist should contain the server socket"
            assert len(d.sockets) == 1, "daemon without connections should have just 1 socket"
        config.SERVERTYPE = old_servertype

    def testServertypeFoobar(self):
        old_servertype = config.SERVERTYPE
        config.SERVERTYPE = "foobar"
        try:
            with pytest.raises(PyroError):
                Pyro5.server.Daemon()
        finally:
            config.SERVERTYPE = old_servertype

    def testRegisterTwice(self):
        with Pyro5.server.Daemon(port=0) as d:
            o1 = MyObj("object1")
            d.register(o1)
            with pytest.raises(DaemonError) as x:
                d.register(o1)
            assert str(x.value) == "object or class already has a Pyro id"
            d.unregister(o1)
            d.register(o1, "samename")
            o2 = MyObj("object2")
            with pytest.raises(DaemonError) as x:
                d.register(o2, "samename")
            assert str(x.value) == "an object or class is already registered with that id"
            assert hasattr(o1, "_pyroId")
            assert hasattr(o1, "_pyroDaemon")
            d.unregister(o1)
            assert not(hasattr(o1, "_pyroId"))
            assert not(hasattr(o1, "_pyroDaemon"))
            o1._pyroId = "FOOBAR"
            d.register(o1)
            o1._pyroId = ""
            d.register(o1)  # with empty-string _pyroId register should work

    def testRegisterInstanceAndClass(self):
        with Pyro5.server.Daemon(port=0) as d:
            o1 = MyObj("object1")
            o2 = MyObj("object2")
            d.register(MyObj)
            d.register(o1)
            d.register(o2)
            with pytest.raises(DaemonError) as x:
                d.register(o2)
            assert str(x.value) == "object or class already has a Pyro id"
            with pytest.raises(DaemonError) as x:
                d.register(MyObj)
            assert str(x.value) == "object or class already has a Pyro id"

    def testRegisterTwiceForced(self):
        with Pyro5.server.Daemon(port=0) as d:
            o1 = MyObj("object1")
            d.register(o1, "name1")
            d.register(o1, "name2", force=True)
            d.register(o1, "name1", force=True)
            assert d.objectsById["name1"] is d.objectsById["name2"]
            d.unregister(o1)
            o1._pyroId = "FOOBAR_ID"
            d.register(o1, "newname", force=True)
            assert o1._pyroId == "newname"
            assert "newname" in d.objectsById

    def testRegisterEtc(self):
        with Pyro5.server.Daemon(port=0) as d:
            assert len(d.objectsById) == 1
            o1 = MyObj("object1")
            o2 = MyObj("object2")
            d.register(o1)
            with pytest.raises(DaemonError):
                d.register(o2, Pyro5.core.DAEMON_NAME)  # cannot use daemon name
            d.register(o2, "obj2a")

            assert len(d.objectsById) == 3
            assert d.objectsById[o1._pyroId] == o1
            assert d.objectsById["obj2a"] == o2
            assert o2._pyroId == "obj2a"
            assert o2._pyroDaemon == d

            # test unregister
            d.unregister("unexisting_thingie")
            with pytest.raises(ValueError):
                d.unregister(None)
            d.unregister("obj2a")
            d.unregister(o1._pyroId)
            assert len(d.objectsById) == 1
            assert o1._pyroId not in d.objectsById
            assert o2._pyroId not in d.objectsById

            # test unregister objects
            del o2._pyroId
            d.register(o2)
            objectid = o2._pyroId
            assert objectid in d.objectsById
            assert len(d.objectsById) == 2
            d.unregister(o2)
            # no more _pyro attributes must remain after unregistering
            for attr in vars(o2):
                assert not(attr.startswith("_pyro"))
            assert len(d.objectsById) == 1
            assert objectid not in d.objectsById
            with pytest.raises(DaemonError):
                d.unregister([1,2,3])

            # test unregister daemon name
            d.unregister(Pyro5.core.DAEMON_NAME)
            assert Pyro5.core.DAEMON_NAME in d.objectsById

            # weird args
            w = MyObj("weird")
            with pytest.raises(AttributeError):
                d.register(None)
            with pytest.raises(AttributeError):
                d.register(4444)
            with pytest.raises(TypeError):
                d.register(w, 666)

            # uri return value from register
            uri = d.register(MyObj("xyz"))
            assert isinstance(uri, Pyro5.core.URI)
            uri = d.register(MyObj("xyz"), "test.register")
            assert uri.object == "test.register"

    def testRegisterClass(self):
        with Pyro5.server.Daemon(port=0) as d:
            assert len(d.objectsById) == 1
            d.register(MyObj)
            with pytest.raises(DaemonError):
                d.register(MyObj)
            assert len(d.objectsById) == 2
            d.uriFor(MyObj)
            # unregister:
            d.unregister(MyObj)
            assert len(d.objectsById) == 1

    def testRegisterUnicode(self):
        with Pyro5.server.Daemon(port=0) as d:
            myobj1 = MyObj("hello1")
            myobj3 = MyObj("hello3")
            uri1 = d.register(myobj1, "str_name")
            uri3 = d.register(myobj3, "unicode_" + chr(0x20ac))
            assert len(d.objectsById) == 3
            uri = d.uriFor(myobj1)
            assert uri == uri1
            _ = Pyro5.client.Proxy(uri)
            uri = d.uriFor(myobj3)
            assert uri == uri3
            _ = Pyro5.client.Proxy(uri)
            uri = d.uriFor("str_name")
            assert uri == uri1
            _ = Pyro5.client.Proxy(uri)
            _ = Pyro5.client.Proxy(uri)
            uri = d.uriFor("unicode_" + chr(0x20ac))
            assert uri == uri3
            _ = Pyro5.client.Proxy(uri)

    def testDaemonObject(self):
        with Pyro5.server.Daemon(port=0) as d:
            daemon = Pyro5.server.DaemonObject(d)
            obj1 = MyObj("object1")
            obj2 = MyObj("object2")
            obj3 = MyObj("object2")
            d.register(obj1, "obj1")
            d.register(obj2, "obj2")
            d.register(obj3)
            daemon.ping()
            registered = daemon.registered()
            assert type(registered) is list
            assert len(registered) == 4
            assert "obj1" in registered
            assert "obj2" in registered
            assert obj3._pyroId in registered
            d.shutdown()

    def testUriFor(self):
        d = Pyro5.server.Daemon(port=0)
        try:
            o1 = MyObj("object1")
            o2 = MyObj("object2")
            with pytest.raises(DaemonError):
                d.uriFor(o1)
            with pytest.raises(DaemonError):
                d.uriFor(o2)
            d.register(o1, None)
            d.register(o2, "object_two")
            o3 = MyObj("object3")
            with pytest.raises(DaemonError):
                d.uriFor(o3)  # can't get an uri for an unregistered object (note: unregistered name is ok)
            u1 = d.uriFor(o1)
            u2 = d.uriFor(o2._pyroId)
            u3 = d.uriFor("unexisting_thingie")  # unregistered name is no problem, it's just an uri we're requesting
            u4 = d.uriFor(o2)
            assert type(u1) == Pyro5.core.URI
            assert u1.protocol == "PYRO"
            assert u2.protocol == "PYRO"
            assert u3.protocol == "PYRO"
            assert u4.protocol == "PYRO"
            assert u4.object == "object_two"
            assert u3 == Pyro5.core.URI("PYRO:unexisting_thingie@" + d.locationStr)
        finally:
            d.close()

    def testDaemonWithStmt(self):
        d = Pyro5.server.Daemon()
        assert d.transportServer
        d.close()  # closes the transportserver and sets it to None
        assert d.transportServer is None
        with Pyro5.server.Daemon() as d:
            assert d.transportServer
            pass
        assert d.transportServer is None
        with pytest.raises(ZeroDivisionError):
            with Pyro5.server.Daemon() as d:
                print(1 // 0)  # cause an error
        assert d.transportServer is None
        d = Pyro5.server.Daemon()
        with d:
            pass
        with pytest.raises(Pyro5.errors.PyroError):
            with d:
                pass
        d.close()

    def testRequestloopCondition(self):
        with Pyro5.server.Daemon(port=0) as d:
            condition = lambda: False
            start = time.time()
            d.requestLoop(loopCondition=condition)  # this should return almost immediately
            duration = time.time() - start
            assert duration < 0.4

    def testSimpleHandshake(self):
        conn = ConnectionMock()
        with Pyro5.server.Daemon(port=0) as d:
            self.sendHandshakeMessage(conn)
            success = d._handshake(conn)
            assert success
            msg = Pyro5.protocol.recv_stub(conn)
            assert msg.type == Pyro5.protocol.MSG_CONNECTOK
            assert msg.seq == 99

    def testHandshakeDenied(self):
        class HandshakeFailDaemon(Pyro5.server.Daemon):
            def validateHandshake(self, conn, data):
                raise ValueError("handshake fail validation error")
        conn = ConnectionMock()
        with HandshakeFailDaemon(port=0) as d:
            self.sendHandshakeMessage(conn)
            success = d._handshake(conn)
            assert not(success)
            msg = Pyro5.protocol.recv_stub(conn)
            assert msg.type == Pyro5.protocol.MSG_CONNECTFAIL
            assert msg.seq == 99
            assert b"handshake fail validation error" in msg.data
        with Pyro5.server.Daemon(port=0) as d:
            self.sendHandshakeMessage(conn)
            success = d._handshake(conn, denied_reason="no way, handshake denied")
            assert not(success)
            msg = Pyro5.protocol.recv_stub(conn)
            assert msg.type == Pyro5.protocol.MSG_CONNECTFAIL
            assert msg.seq == 99
            assert b"no way, handshake denied" in msg.data

    def testCustomHandshake(self):
        conn = ConnectionMock()
        class CustomHandshakeDaemon(Pyro5.server.Daemon):
            def validateHandshake(self, conn, data):
                return ["sure", "have", "fun"]
            def annotations(self):
                return {"XYZZ": b"custom annotation set by daemon"}
        with CustomHandshakeDaemon(port=0) as d:
            corr_id = uuid.uuid4()
            self.sendHandshakeMessage(conn, correlation_id=corr_id)
            assert current_context.correlation_id == corr_id
            success = d._handshake(conn)
            assert success
            msg = Pyro5.protocol.recv_stub(conn)
            assert msg.type == Pyro5.protocol.MSG_CONNECTOK
            assert msg.seq == 99
            assert len(msg.annotations) == 1
            assert msg.annotations["XYZZ"] == b"custom annotation set by daemon"
            ser = Pyro5.serializers.serializers_by_id[msg.serializer_id]
            data = ser.loads(msg.data)
            assert data["handshake"] == ["sure", "have", "fun"]

    def testNAT(self):
        with Pyro5.server.Daemon() as d:
            assert d.natLocationStr is None
        with Pyro5.server.Daemon(nathost="nathosttest", natport=12345) as d:
            assert d.natLocationStr == "nathosttest:12345"
            assert d.natLocationStr != d.locationStr
            uri = d.register(MyObj(1))
            assert uri.location == "nathosttest:12345"
            uri = d.uriFor("object")
            assert uri.location == "nathosttest:12345"
            uri = d.uriFor("object", nat=False)
            assert uri.location != "nathosttest:12345"
            d = Pyro5.server.Daemon(nathost="bla")
            assert d.natLocationStr.startswith("bla:")
        with pytest.raises(ValueError):
            Pyro5.server.Daemon(natport=5555)
        with pytest.raises(ValueError):
            Pyro5.server.Daemon(nathost="bla", natport=5555, unixsocket="testsock")

    def testNATzeroPort(self):
        servertype = config.SERVERTYPE
        try:
            config.SERVERTYPE = "multiplex"
            with Pyro5.server.Daemon(nathost="nathosttest", natport=99999) as d:
                host, port = d.locationStr.split(":")
                assert port != 99999
                assert d.natLocationStr == "nathosttest:99999"
            with Pyro5.server.Daemon(nathost="nathosttest", natport=0) as d:
                host, port = d.locationStr.split(":")
                assert d.natLocationStr == "nathosttest:%s" % port
            config.SERVERTYPE = "thread"
            with Pyro5.server.Daemon(nathost="nathosttest", natport=99999) as d:
                host, port = d.locationStr.split(":")
                assert port != 99999
                assert d.natLocationStr == "nathosttest:99999"
            with Pyro5.server.Daemon(nathost="nathosttest", natport=0) as d:
                host, port = d.locationStr.split(":")
                assert d.natLocationStr == "nathosttest:%s" % port
        finally:
            config.SERVERTYPE = servertype

    def testNATconfig(self):
        try:
            config.NATHOST = None
            config.NATPORT = 0
            with Pyro5.server.Daemon() as d:
                assert d.natLocationStr is None
            config.NATHOST = "nathosttest"
            config.NATPORT = 12345
            with Pyro5.server.Daemon() as d:
                assert d.natLocationStr == "nathosttest:12345"
        finally:
            config.NATHOST = None
            config.NATPORT = 0

    def testBehaviorDefaults(self):
        class TestClass:
            pass
        with Pyro5.server.Daemon() as d:
            d.register(TestClass)
            instance_mode, instance_creator = TestClass._pyroInstancing
            assert instance_mode == "session"
            assert instance_creator is None

    def testInstanceCreationSingle(self):
        def creator(clazz):
            return clazz("testname")
        @Pyro5.server.behavior(instance_mode="single", instance_creator=creator)
        class TestClass:
            def __init__(self, name):
                self.name = name
        conn = Pyro5.socketutil.SocketConnection(socket.socket())
        d = Pyro5.server.Daemon()
        instance1 = d._getInstance(TestClass, conn)
        instance2 = d._getInstance(TestClass, conn)
        assert instance1.name == "testname"
        assert instance1 is instance2
        assert TestClass in d._pyroInstances
        assert instance1 is d._pyroInstances[TestClass]
        assert not(TestClass in conn.pyroInstances)

    def testBehaviorDefaultsIsSession(self):
        class ClassWithDefaults:
            def __init__(self):
                self.name = "yep"
        conn1 = Pyro5.socketutil.SocketConnection(socket.socket())
        conn2 = Pyro5.socketutil.SocketConnection(socket.socket())
        d = Pyro5.server.Daemon()
        d.register(ClassWithDefaults)
        instance1a = d._getInstance(ClassWithDefaults, conn1)
        instance1b = d._getInstance(ClassWithDefaults, conn1)
        instance2a = d._getInstance(ClassWithDefaults, conn2)
        instance2b = d._getInstance(ClassWithDefaults, conn2)
        assert instance1a is instance1b
        assert instance2a is instance2b
        assert instance1a is not instance2a
        assert not(ClassWithDefaults in d._pyroInstances)
        assert ClassWithDefaults in conn1.pyroInstances
        assert ClassWithDefaults in conn2.pyroInstances
        assert instance1a is conn1.pyroInstances[ClassWithDefaults]
        assert instance2a is conn2.pyroInstances[ClassWithDefaults]

    def testInstanceCreationSession(self):
        def creator(clazz):
            return clazz("testname")

        @Pyro5.server.behavior(instance_mode="session", instance_creator=creator)
        class ClassWithDecorator:
            def __init__(self, name):
                self.name = name
        conn1 = Pyro5.socketutil.SocketConnection(socket.socket())
        conn2 = Pyro5.socketutil.SocketConnection(socket.socket())
        d = Pyro5.server.Daemon()
        d.register(ClassWithDecorator)
        # check the class with the decorator first
        instance1a = d._getInstance(ClassWithDecorator, conn1)
        instance1b = d._getInstance(ClassWithDecorator, conn1)
        instance2a = d._getInstance(ClassWithDecorator, conn2)
        instance2b = d._getInstance(ClassWithDecorator, conn2)
        assert instance1a is instance1b
        assert instance2a is instance2b
        assert instance1a is not instance2a
        assert not(ClassWithDecorator in d._pyroInstances)
        assert ClassWithDecorator in conn1.pyroInstances
        assert ClassWithDecorator in conn2.pyroInstances
        assert instance1a is conn1.pyroInstances[ClassWithDecorator]
        assert instance2a is conn2.pyroInstances[ClassWithDecorator]


    def testInstanceCreationPerCall(self):
        def creator(clazz):
            return clazz("testname")
        @Pyro5.server.behavior(instance_mode="percall", instance_creator=creator)
        class TestClass:
            def __init__(self, name):
                self.name = name
        with Pyro5.socketutil.SocketConnection(socket.socket()) as conn:
            with Pyro5.server.Daemon() as d:
                instance1 = d._getInstance(TestClass, conn)
                instance2 = d._getInstance(TestClass, conn)
                assert instance1 is not instance2
                assert not(TestClass in d._pyroInstances)
                assert not(TestClass in conn.pyroInstances)

    def testInstanceCreationWrongType(self):
        def creator(clazz):
            return Pyro5.core.URI("PYRO:test@localhost:9999")
        @Pyro5.server.behavior(instance_creator=creator)
        class TestClass:
            def method(self):
                pass
        with Pyro5.socketutil.SocketConnection(socket.socket()) as conn:
            with Pyro5.server.Daemon() as d:
                with pytest.raises(TypeError):
                    d._getInstance(TestClass, conn)

    def testCombine(self):
        d1 = Pyro5.server.Daemon()
        d2 = Pyro5.server.Daemon()
        with pytest.raises(TypeError):
            d1.combine(d2)
        d1.close()
        d2.close()
        try:
            config.SERVERTYPE = "multiplex"
            d1 = Pyro5.server.Daemon()
            d2 = Pyro5.server.Daemon()
            nsuri, nsd, bcd = Pyro5.nameserver.start_ns(host="", bchost="")
            d1_selector = d1.transportServer.selector
            d1.combine(d2)
            d1.combine(nsd)
            d1.combine(bcd)
            assert d1_selector is d1.transportServer.selector
            assert d1_selector is d2.transportServer.selector
            assert d1_selector is nsd.transportServer.selector
            assert d1_selector is bcd.transportServer.selector
            assert len(d1.sockets) == 4
            assert d1.sock in d1.sockets
            assert d2.sock in d1.sockets
            assert nsd.sock in d1.sockets
            assert bcd in d1.sockets
            bcd.close()
            nsd.close()
            d2.close()
            d1.close()
        finally:
            config.SERVERTYPE = "thread"


class TestMetaInfo:
    def testMeta(self):
        with Pyro5.server.Daemon() as d:
            daemon_obj = d.objectsById[Pyro5.core.DAEMON_NAME]
            assert len(daemon_obj.info()) > 10
            meta = daemon_obj.get_metadata(Pyro5.core.DAEMON_NAME)
            assert meta["methods"] == {"get_metadata", "get_next_stream_item", "close_stream", "info", "ping", "registered"}

    def testMetaSerialization(self):
        with Pyro5.server.Daemon() as d:
            daemon_obj = d.objectsById[Pyro5.core.DAEMON_NAME]
            meta = daemon_obj.get_metadata(Pyro5.core.DAEMON_NAME)
            for ser_id in [Pyro5.serializers.JsonSerializer.serializer_id,
                           Pyro5.serializers.MarshalSerializer.serializer_id,
                           Pyro5.serializers.SerpentSerializer.serializer_id]:
                serializer = Pyro5.serializers.serializers_by_id[ser_id]
                data = serializer.dumps(meta)
                _ = serializer.loads(data)

    def testMetaResetCache(self):
        class Dummy:
            @Pyro5.server.expose
            def method(self):
                pass
        with Pyro5.server.Daemon() as d:
            dummy = Dummy()
            uri = d.register(dummy)
            daemon_obj = d.objectsById[Pyro5.core.DAEMON_NAME]
            meta = daemon_obj.get_metadata(uri.object)
            assert "newly_added_method" not in meta["methods"]
            assert "newly_added_method_two" not in meta["methods"]
            Dummy.newly_added_method = Pyro5.server.expose(lambda self: None)
            meta = daemon_obj.get_metadata(uri.object)
            assert "newly_added_method" not in meta["methods"]
            d.resetMetadataCache(uri.object)
            meta = daemon_obj.get_metadata(uri.object)
            assert "newly_added_method" in meta["methods"]
            Dummy.newly_added_method_two = Pyro5.server.expose(lambda self: None)
            d.resetMetadataCache(dummy)
            meta = daemon_obj.get_metadata(uri.object)
            assert "newly_added_method_two" in meta["methods"]
            del Dummy.newly_added_method
            del Dummy.newly_added_method_two



import zlib
import pytest
import Pyro5.protocol
import Pyro5.errors
import Pyro5.protocol
import Pyro5.serializers
import Pyro5.errors
from Pyro5.protocol import SendingMessage, ReceivingMessage
from support import ConnectionMock


class TestSendingMessage:
    def test_create(self):
        msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, 99, b"abcdefg")
        assert len(msg.data) > 1

    def test_annotations_errors(self):
        with pytest.raises(Pyro5.errors.ProtocolError):
            Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, 99, b"abcdefg", annotations={"zxcv": "no_bytes"})
        with pytest.raises(Pyro5.errors.ProtocolError):
            Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, 99, b"abcdefg", annotations={"err": b"bytes"})

    def test_annotations(self):
        msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, 99, b"abcdefg", annotations={"zxcv": b"bytes"})
        assert len(msg.data) > 1

    def test_compression(self):
        compr_orig = Pyro5.config.COMPRESSION
        try:
            Pyro5.config.COMPRESSION = False
            msg_uncompressed = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, 99, b"abcdefg"*100)
            Pyro5.config.COMPRESSION = True
            msg_compressed = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, 99, b"abcdefg"*100)
            msg_notcompressed = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, 99, b"abcdefg")
            assert not (msg_uncompressed.flags & Pyro5.protocol.FLAGS_COMPRESSED)
            assert not (msg_notcompressed.flags & Pyro5.protocol.FLAGS_COMPRESSED)
            assert msg_compressed.flags & Pyro5.protocol.FLAGS_COMPRESSED
            assert len(msg_uncompressed.data) > len(msg_compressed.data)
        finally:
            Pyro5.config.COMPRESSION = compr_orig


class TestReceivingMessage:
    def createmessage(self, compression=False):
        compr_orig = Pyro5.config.COMPRESSION
        Pyro5.config.COMPRESSION = compression
        annotations = {
            "TEST": b"0123456789qwertyu",
            "UNIT": b"aaaaaaa"
        }
        msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 255, 42, 99, b"abcdefg"*100, annotations)
        Pyro5.config.COMPRESSION = compr_orig
        return msg

    def test_validate(self):
        with pytest.raises(ValueError):
            Pyro5.protocol.ReceivingMessage.validate(b"abc")
        with pytest.raises(Pyro5.errors.ProtocolError):
            Pyro5.protocol.ReceivingMessage.validate(b"ZXCV")
        Pyro5.protocol.ReceivingMessage.validate(b"PYRO")
        with pytest.raises(Pyro5.errors.ProtocolError):
            Pyro5.protocol.ReceivingMessage.validate(b"PYRO__")
        msg = self.createmessage()
        msg.data = bytearray(msg.data)
        Pyro5.protocol.ReceivingMessage.validate(msg.data)
        orig_magic = msg.data[38]
        msg.data[38] = 0xff   # kill the magic number
        with pytest.raises(Pyro5.errors.ProtocolError) as x:
            Pyro5.protocol.ReceivingMessage.validate(msg.data)
        assert "magic number" in str(x.value)
        msg.data[38] = orig_magic   # repair the magic number
        msg.data[5] = 0xff   # invalid protocol version
        with pytest.raises(Pyro5.errors.ProtocolError) as x:
            Pyro5.protocol.ReceivingMessage.validate(msg.data)
        assert "protocol version" in str(x.value)

    def test_create_nopayload(self):
        send_msg = self.createmessage(compression=True)
        header = send_msg.data[:Pyro5.protocol._header_size]
        msg = Pyro5.protocol.ReceivingMessage(header)
        assert msg.data is None
        assert msg.type == Pyro5.protocol.MSG_INVOKE
        assert msg.flags == 255
        assert msg.seq == 42
        assert msg.serializer_id == 99
        assert len(msg.annotations) == 0

    def test_create_payload(self):
        send_msg = self.createmessage(compression=True)
        header = send_msg.data[:Pyro5.protocol._header_size]
        payload = send_msg.data[Pyro5.protocol._header_size:]
        msg = Pyro5.protocol.ReceivingMessage(header, payload)
        assert len(msg.data) == 700
        assert msg.flags == 255 & ~Pyro5.protocol.FLAGS_COMPRESSED
        assert len(msg.annotations) == 2
        assert msg.annotations["TEST"] == b"0123456789qwertyu"
        assert msg.annotations["UNIT"] == b"aaaaaaa"
        assert type(msg.data) is bytes

    def test_create_payload_memview(self):
        send_msg = self.createmessage(compression=False)
        header = send_msg.data[:Pyro5.protocol._header_size]
        payload = send_msg.data[Pyro5.protocol._header_size:]
        msg = Pyro5.protocol.ReceivingMessage(header, payload)
        assert type(msg.data) is memoryview


class TestProtocolMessages:
    def testMessage(self):
        SendingMessage(99, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, b"")  # doesn't check msg type here
        with pytest.raises(TypeError):
            ReceivingMessage("FOOBAR")
        with pytest.raises(Pyro5.errors.ProtocolError):
            ReceivingMessage(b"FOOBARFOOBARFOOBARFOOBARFOOBARFOOBAR1234")
        msg = SendingMessage(Pyro5.protocol.MSG_CONNECT, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, b"hello")
        assert msg.type == Pyro5.protocol.MSG_CONNECT
        assert len(msg.data) == Pyro5.protocol._header_size + 5
        msg = ReceivingMessage(msg.data[:Pyro5.protocol._header_size], msg.data[Pyro5.protocol._header_size:])
        assert msg.type == Pyro5.protocol.MSG_CONNECT
        assert msg.data == b"hello"

        msg = SendingMessage(Pyro5.protocol.MSG_RESULT, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, b"")
        msg = ReceivingMessage(msg.data[:Pyro5.protocol._header_size], msg.data[Pyro5.protocol._header_size:])
        assert msg.type == Pyro5.protocol.MSG_RESULT
        assert len(msg.data) == 0

        msg = SendingMessage(255, 0, 255, Pyro5.serializers.SerpentSerializer.serializer_id, b"").data
        assert len(msg) == 40
        msg = SendingMessage(1, 0, 255, Pyro5.serializers.SerpentSerializer.serializer_id, b"").data
        assert len(msg) == 40

        data = b"x" * 1000
        Pyro5.config.COMPRESSION = True
        msg = SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, data).data
        assert len(msg) < len(data)
        Pyro5.config.COMPRESSION = False

    def testAnnotationsIdLength4(self):
        with pytest.raises(Pyro5.errors.ProtocolError):
            SendingMessage(Pyro5.protocol.MSG_CONNECT, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, b"hello", {"TOOLONG": b"abcde"})
        with pytest.raises(Pyro5.errors.ProtocolError):
            SendingMessage(Pyro5.protocol.MSG_CONNECT, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, b"hello", {"QQ": b"abcde"})

    def testRecvAnnotations(self):
        annotations = {"TEST": b"abcde"}
        msg = SendingMessage(Pyro5.protocol.MSG_CONNECT, 0, 0, Pyro5.serializers.SerpentSerializer.serializer_id, b"hello", annotations)
        c = ConnectionMock()
        c.send(msg.data)
        msg = Pyro5.protocol.recv_stub(c)
        assert len(c.received) == 0
        assert msg.data == b"hello"
        assert msg.annotations["TEST"] == b"abcde"

    def testCompression(self):
        data = b"The quick brown fox jumps over the lazy dog."*10
        compressed_data = zlib.compress(data)
        flags = Pyro5.protocol.FLAGS_COMPRESSED
        msg = SendingMessage(Pyro5.protocol.MSG_INVOKE, 42, flags, 1, compressed_data)
        assert msg.data != data
        assert len(msg.data) < len(data)

    def testRecvNoAnnotations(self):
        msg = SendingMessage(Pyro5.protocol.MSG_CONNECT, 42, 0, 0, b"hello")
        c = ConnectionMock()
        c.send(msg.data)
        msg = Pyro5.protocol.recv_stub(c)
        assert len(c.received) == 0
        assert msg.data_size == 5
        assert msg.data == b"hello"
        assert msg.annotations_size == 0
        assert len(msg.annotations) == 0
